/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#ifndef __LIS_SOLVER_H__
#define __LIS_SOLVER_H__

typedef int (*LIS_SOLVER_CHECK_PARAMS)(LIS_SOLVER solver);
typedef int (*LIS_SOLVER_MALLOC_WORK)(LIS_SOLVER solver);
typedef int (*LIS_SOLVER_EXECUTE)(LIS_SOLVER solver);

#ifdef __cplusplus
extern "C"
{
#endif
	extern int lis_solver_init(LIS_SOLVER solver);
	extern int lis_solver_work_destroy(LIS_SOLVER solver);
	extern int lis_solver_get_initial_residual(LIS_SOLVER solver, LIS_PRECON M, LIS_VECTOR t, LIS_VECTOR r, LIS_SCALAR *bnrm2);
	extern int lis_solver_set_option2(char* arg1, char *arg2, LIS_SOLVER solver);
	extern int lis_solver_set_option_solver(char *argv, LIS_SOLVER solver);
	extern int lis_solver_set_option_psolver(char *argv, LIS_SOLVER solver);
	extern int lis_solver_set_option_precon(char *argv, LIS_SOLVER solver);
	extern int lis_solver_set_option_pprecon(char *argv, LIS_SOLVER solver);
	extern int lis_solver_set_option_scale(char *argv, LIS_SOLVER solver);
	extern int lis_solver_set_option_print(char *argv, LIS_SOLVER solver);
	extern int lis_solver_set_option_truefalse(char *argv, int opt, LIS_SOLVER solver);
	extern int lis_solver_set_option_precision(char *argv, int opt, LIS_SOLVER solver);
	extern int lis_solver_set_option_storage(char *argv, LIS_SOLVER solver);
	/*******************/
	/* CG              */
	/*******************/
	extern int lis_cg(LIS_SOLVER solver);
	extern int lis_cg_quad(LIS_SOLVER solver);
	extern int lis_cg_switch(LIS_SOLVER solver);
	extern int lis_cg_periodic(LIS_SOLVER solver);
	extern int lis_cg_check_params(LIS_SOLVER solver);
	extern int lis_cg_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* BiCG            */
	/*******************/
	extern int lis_bicg(LIS_SOLVER solver);
	extern int lis_bicg_quad(LIS_SOLVER solver);
	extern int lis_bicg_periodic(LIS_SOLVER solver);
	extern int lis_bicg_switch(LIS_SOLVER solver);
	extern int lis_bicg_check_params(LIS_SOLVER solver);
	extern int lis_bicg_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* CGS             */
	/*******************/
	extern int lis_cgs(LIS_SOLVER solver);
	extern int lis_cgs_quad(LIS_SOLVER solver);
	extern int lis_cgs_periodic(LIS_SOLVER solver);
	extern int lis_cgs_switch(LIS_SOLVER solver);
	extern int lis_cgs_check_params(LIS_SOLVER solver);
	extern int lis_cgs_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* BiCGSTAB        */
	/*******************/
	extern int lis_bicgstab(LIS_SOLVER solver);
	extern int lis_bicgstab_quad(LIS_SOLVER solver);
	extern int lis_bicgstab_periodic(LIS_SOLVER solver);
	extern int lis_bicgstab_switch(LIS_SOLVER solver);
	extern int lis_bicgstab_check_params(LIS_SOLVER solver);
	extern int lis_bicgstab_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* BiCGSTAB(l)     */
	/*******************/
	extern int lis_bicgstabl(LIS_SOLVER solver);
	extern int lis_bicgstabl_quad(LIS_SOLVER solver);
	extern int lis_bicgstabl_check_params(LIS_SOLVER solver);
	extern int lis_bicgstabl_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* GPBiCG          */
	/*******************/
	extern int lis_gpbicg(LIS_SOLVER solver);
	extern int lis_gpbicg_quad(LIS_SOLVER solver);
	extern int lis_gpbicg_periodic(LIS_SOLVER solver);
	extern int lis_gpbicg_switch(LIS_SOLVER solver);
	extern int lis_gpbicg_check_params(LIS_SOLVER solver);
	extern int lis_gpbicg_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* BiCGSafe        */
	/*******************/
	extern int lis_bicgsafe(LIS_SOLVER solver);
	extern int lis_bicgsafe_quad(LIS_SOLVER solver);
	extern int lis_bicgsafe_switch(LIS_SOLVER solver);
	extern int lis_bicgsafe_check_params(LIS_SOLVER solver);
	extern int lis_bicgsafe_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* TFQMR           */
	/*******************/
	extern int lis_tfqmr(LIS_SOLVER solver);
	extern int lis_tfqmr_quad(LIS_SOLVER solver);
	extern int lis_tfqmr_check_params(LIS_SOLVER solver);
	extern int lis_tfqmr_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* Orthomin(m)     */
	/*******************/
	extern int lis_orthomin(LIS_SOLVER solver);
	extern int lis_orthomin_quad(LIS_SOLVER solver);
	extern int lis_orthomin_check_params(LIS_SOLVER solver);
	extern int lis_orthomin_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* GMRES(m)        */
	/*******************/
	extern int lis_gmres(LIS_SOLVER solver);
	extern int lis_gmres_quad(LIS_SOLVER solver);
	extern int lis_gmres_switch(LIS_SOLVER solver);
	extern int lis_gmres_check_params(LIS_SOLVER solver);
	extern int lis_gmres_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* Jacobi          */
	/*******************/
	extern int lis_jacobi(LIS_SOLVER solver);
	extern int lis_jacobi_check_params(LIS_SOLVER solver);
	extern int lis_jacobi_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* Gauss-Seidel    */
	/*******************/
	extern int lis_gs(LIS_SOLVER solver);
	extern int lis_gs_check_params(LIS_SOLVER solver);
	extern int lis_gs_malloc_work(LIS_SOLVER solver);
	/*******************/
	/* SOR             */
	/*******************/
	extern int lis_sor(LIS_SOLVER solver);
	extern int lis_sor_check_params(LIS_SOLVER solver);
	extern int lis_sor_malloc_work(LIS_SOLVER solver);

#ifdef __cplusplus
}
#endif
#endif
