/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_solver_create
 * lis_solver_destroy
 * lis_solver_set_option
 * lis_solver_get_option
 * lis_solve
 ************************************************/
#ifdef USE_FORTRAN

#undef __FUNC__
#define __FUNC__ "lis_solver_create_f"
void lis_solver_create_f(LIS_SOLVER_F *solver, int *ierr)
{
	LIS_SOLVER	s;

	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_create(&s);
	if( *ierr )	return;

	*solver = LIS_P2V(s);
	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_destroy_f"
void lis_solver_destroy_f(LIS_SOLVER_F *solver, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_destroy((LIS_SOLVER)LIS_V2P(solver));
	if( *ierr )	return;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solve_f"
void lis_solve_f(LIS_MATRIX_F *A, LIS_VECTOR_F b, LIS_VECTOR_F x, LIS_SOLVER_F *solver, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solve((LIS_MATRIX)LIS_V2P(A), (LIS_VECTOR)LIS_V2P(b), (LIS_VECTOR)LIS_V2P(x), (LIS_SOLVER)LIS_V2P(solver));
	if( *ierr )	return;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_f"
void lis_solver_set_option_f(char *text, LIS_SOLVER_F *solver, int *ierr, int len)
{
	char	buf[1024];
	LIS_DEBUG_FUNC_IN;

	strncpy(buf,text,len);
	buf[len] = '\0';
	*ierr = lis_solver_set_option(buf,(LIS_SOLVER)LIS_V2P(solver));
	if( *ierr )	return;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_optionC_f"
void lis_solver_set_optionC_f(LIS_SOLVER solver, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_set_optionC((LIS_SOLVER)LIS_V2P(solver));

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_iters_f"
void lis_solver_get_iters_f(LIS_SOLVER_F *solver, int *iters, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_get_iters((LIS_SOLVER)LIS_V2P(solver),iters);
	if( *ierr )	return;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_times_f"
void lis_solver_get_times_f(LIS_SOLVER_F *solver, double *times, double *itimes, double *ptimes, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_get_times((LIS_SOLVER)LIS_V2P(solver),times,itimes,ptimes);
	if( *ierr )	return;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_residualnorm_f"
void lis_solver_get_residualnorm_f(LIS_SOLVER_F *solver, LIS_REAL *residual, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_get_residualnorm((LIS_SOLVER)LIS_V2P(solver),residual);
	if( *ierr )	return;

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_solver_f"
void lis_solver_get_solver_f(LIS_SOLVER_F *solver, int *nsol, int *ierr)
{
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_solver_get_solver((LIS_SOLVER)LIS_V2P(solver),nsol);

	LIS_DEBUG_FUNC_OUT;
	return;
}

#undef __FUNC__
#define __FUNC__ "lis_get_solvername_f"
void lis_get_solvername_f(int *solver, char *name, int *ierr, int len)
{
	char	buf[1024];
	LIS_DEBUG_FUNC_IN;

	*ierr = lis_get_solvername(*solver, buf);
	strncpy(name,buf,len);

	LIS_DEBUG_FUNC_OUT;
	return;
}

#endif
