/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

#undef __FUNC__
#define __FUNC__ "lis_matrix_scaling"
int lis_matrix_scaling(LIS_MATRIX A, LIS_VECTOR B, LIS_VECTOR D, int action)
{
	int			i,n,np;
	LIS_SCALAR  *b,*d;

	n  = A->n;
	np = A->np;
	b  = B->value;
	d  = D->value;

	lis_matrix_get_diagonal(A,D);
	if( action==LIS_SCALE_SYMM_DIAG )
	{
#ifdef USE_MPI
		if( A->np>D->np )
		{
			D->value = (LIS_SCALAR *)lis_realloc(D->value,A->np*sizeof(LIS_SCALAR));
			if( D->value==NULL )
			{
				LIS_SETERR_MEM(A->np*sizeof(LIS_SCALAR));
				return LIS_OUT_OF_MEMORY;
			}
			d = D->value;
		}
		lis_send_recv(A->commtable,d);
#endif
#pragma omp parallel for private(i)
		for(i=0; i<np; i++)
		{
			d[i] = 1.0 / sqrt(fabs(d[i]));
		}

		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			lis_matrix_scaling_symm_crs(A, d);
			break;
		case LIS_MATRIX_CCS:
			lis_matrix_scaling_symm_ccs(A, d);
			break;
		case LIS_MATRIX_MSR:
			lis_matrix_scaling_symm_msr(A, d);
			break;
		case LIS_MATRIX_DIA:
			lis_matrix_scaling_symm_dia(A, d);
			break;
		case LIS_MATRIX_ELL:
			lis_matrix_scaling_symm_ell(A, d);
			break;
		case LIS_MATRIX_JDS:
			lis_matrix_scaling_symm_jds(A, d);
			break;
		case LIS_MATRIX_BSR:
			lis_matrix_scaling_symm_bsr(A, d);
			break;
		case LIS_MATRIX_BSC:
			lis_matrix_scaling_symm_bsc(A, d);
			break;
		case LIS_MATRIX_DNS:
			lis_matrix_scaling_symm_dns(A, d);
			break;
		case LIS_MATRIX_COO:
			lis_matrix_scaling_symm_coo(A, d);
			break;
/*
		case LIS_MATRIX_VBR:
			lis_matrix_scaling_symm_vbr(A, d);
			break;
*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}
	else
	{
#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			d[i] = 1.0 / d[i];
		}
		switch( A->matrix_type )
		{
		case LIS_MATRIX_CRS:
			lis_matrix_scaling_crs(A, d);
			break;
		case LIS_MATRIX_CCS:
			lis_matrix_scaling_ccs(A, d);
			break;
		case LIS_MATRIX_MSR:
			lis_matrix_scaling_msr(A, d);
			break;
		case LIS_MATRIX_DIA:
			lis_matrix_scaling_dia(A, d);
			break;
		case LIS_MATRIX_ELL:
			lis_matrix_scaling_ell(A, d);
			break;
		case LIS_MATRIX_JDS:
			lis_matrix_scaling_jds(A, d);
			break;
		case LIS_MATRIX_BSR:
			lis_matrix_scaling_bsr(A, d);
			break;
		case LIS_MATRIX_BSC:
			lis_matrix_scaling_bsc(A, d);
			break;
		case LIS_MATRIX_DNS:
			lis_matrix_scaling_dns(A, d);
			break;
		case LIS_MATRIX_COO:
			lis_matrix_scaling_coo(A, d);
			break;
/*
		case LIS_MATRIX_VBR:
			lis_matrix_scaling_vbr(A, d);
			break;
*/
		default:
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
			break;
		}
	}

#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		b[i] = b[i]*d[i];
	}
	A->is_scaled = LIS_TRUE;
	B->is_scaled = LIS_TRUE;
	return LIS_SUCCESS;
}

