/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"



#undef __FUNC__
#define __FUNC__ "lis_matrix_set_cdia"
int lis_matrix_set_cdia(int bnr, int maxnnd, int *nnd, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;
	int nr;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	nr        = (A->n-1)/bnr + 1;

	A->matrix_type = -LIS_MATRIX_CDIA;
	A->bnr         = bnr;
	A->nr          = nr;
	A->nnd         = maxnnd;
	A->col         = nnd;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/*
#undef __FUNC__
#define __FUNC__ "lis_matrix_set_jds"
#ifdef _OPENMP
int lis_matrix_set_jds(int nnz, int *maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	int maxmaxnzr,i,nprocs;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	maxmaxnzr = 0;
	nprocs    = omp_get_max_threads();
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
	}

	A->matrix_type = -LIS_MATRIX_JDS;
	A->nnz         = nnz;
	A->maxnzr      = maxmaxnzr;
	A->col         = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_matrix_set_jds(int nnz, int maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_JDS;
	A->nnz         = nnz;
	A->maxnzr      = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif
*/

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_pcrs"
int lis_matrix_set_pcrs(int nnz, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_PCRS;
	A->nnz         = nnz;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_ljds"
#ifdef _OPENMP
int lis_matrix_set_ljds(int nnz, int *maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;
	int maxmaxnzr,i,nprocs;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	maxmaxnzr = 0;
	nprocs    = omp_get_max_threads();
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
	}

	A->matrix_type = -LIS_MATRIX_JDS;
	A->nnz         = nnz;
	A->maxnzr      = maxmaxnzr;
	A->col         = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_matrix_set_ljds(int nnz, int maxnzr, int nl, int *lptr, int *lnz, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_LJDS;
	A->nnz         = nnz;
	A->maxnzr      = maxnzr;
	A->nl          = nl;
	A->lptr        = lptr;
	A->lnz         = lnz;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_lcrs"
int lis_matrix_set_lcrs(int nnz, int nl, int *lptr, int *lnz, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_LCRS;
	A->nnz         = nnz;
	A->nl          = nl;
	A->lptr        = lptr;
	A->lnz         = lnz;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_cjds"
int lis_matrix_set_cjds(int bnr, int nnz, int *maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;
	int maxmaxnzr,i,nprocs,nr;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	nr        = (A->n-1)/bnr + 1;
	maxmaxnzr = 0;
	for(i=0;i<nr;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
	}

	A->matrix_type = -LIS_MATRIX_CJDS;
	A->bnr         = bnr;
	A->nr          = nr;
	A->nnz         = nnz;
	A->maxnzr      = maxmaxnzr;
	A->col         = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_tjds"
#ifdef _OPENMP
int lis_matrix_set_tjds(int nnz, int *maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;
	int maxmaxnzr,i,nprocs;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	maxmaxnzr = 0;
	nprocs    = omp_get_max_threads();
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
	}

	A->matrix_type = -LIS_MATRIX_TJDS;
	A->nnz         = nnz;
	A->maxnzr      = maxmaxnzr;
	A->col         = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_matrix_set_tjds(int nnz, int maxnzr, int *perm, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_TJDS;
	A->nnz         = nnz;
	A->maxnzr      = maxnzr;
	A->row         = perm;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_bjds"
#ifdef _OPENMP
int lis_matrix_set_bjds(int bnr, int bnc, int bnnz, int *maxnzr, int *perm, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;
	int maxmaxnzr,i,nprocs;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	maxmaxnzr = 0;
	nprocs    = omp_get_max_threads();
	for(i=0;i<nprocs;i++)
	{
		if( maxnzr[i] > maxmaxnzr ) maxmaxnzr = maxnzr[i];
	}

	A->matrix_type = -LIS_MATRIX_BJDS;
	A->block       = LIS_TRUE;
	A->bnnz        = bnnz;
	A->nr          = (A->n-1)/bnr+1;
	A->nc          = (A->gn-1)/bnc+1;
	A->maxnzr      = maxmaxnzr;
	A->bnr         = bnr;
	A->bnc         = bnc;
	A->col         = maxnzr;
	A->row         = perm;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_matrix_set_bjds(int bnr, int bnc, int bnnz, int maxnzr, int *perm, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_BJDS;
	A->block       = LIS_TRUE;
	A->bnnz        = bnnz;
	A->nr          = (A->n-1)/bnr+1;
	A->nc          = (A->gn-1)/bnc+1;
	A->maxnzr      = maxnzr;
	A->bnr         = bnr;
	A->bnc         = bnc;
	A->row         = perm;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;
	A->np          = A->n + (bnr - A->n%bnr)%bnr;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_matrix_set_bcr"
int lis_matrix_set_bcr(int bnr, int bnc, int bnnz, int *cptr, int *ptr, int *index, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_BCR;
	A->block       = LIS_TRUE;
	A->bnnz        = bnnz;
	A->nr          = (A->n-1)/bnr+1;
	A->nc          = (A->gn-1)/bnc+1;
	A->bnr         = bnr;
	A->bnc         = bnc;
	A->cptr        = cptr;
	A->ptr         = ptr;
	A->index       = index;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_matrix_set_lbsr"
int lis_matrix_set_lbsr(int bnr, int bnc, int bnnz, int nl, int *lptr, int *lnz, int *row, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_LBSR;
	A->block       = LIS_TRUE;
	A->bnnz        = bnnz;
	A->nr          = (A->n-1)/bnr+1;
	A->nc          = (A->gn-1)/bnc+1;
	A->bnr         = bnr;
	A->bnc         = bnc;
	A->nl          = nl;
	A->lptr        = lptr;
	A->lnz         = lnz;
	A->row         = row;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;
	A->np          = A->n + (bnr - A->n%bnr)%bnr;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


/*
#undef __FUNC__
#define __FUNC__ "lis_matrix_set_vbr"
int lis_matrix_set_vbr(int nnz, int nr, int nc, int bnnz, int *row, int *col, int *ptr, int *bptr, int *bindex, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_VBR;
	A->block       = LIS_TRUE;
	A->nnz         = nnz;
	A->nr          = nr;
	A->nc          = nc;
	A->bnnz        = bnnz;
	A->row         = row;
	A->col         = col;
	A->ptr         = ptr;
	A->bptr        = bptr;
	A->bindex      = bindex;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
*/
/*
#undef __FUNC__
#define __FUNC__ "lis_matrix_set_coo"
int lis_matrix_set_coo(int nnz, int *row, int *col, LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_COO;
	A->nnz         = nnz;
	A->row         = row;
	A->col         = col;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
*/
/*
#undef __FUNC__
#define __FUNC__ "lis_matrix_set_dns"
int lis_matrix_set_dns(LIS_SCALAR *value, LIS_MATRIX A)
{
	int err;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_TYPE);
	if( err ) return err;

	A->matrix_type = -LIS_MATRIX_DNS;
	A->value       = value;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
*/
