/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_coo(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int n,nnz;

	n    = A->n;
	nnz  = A->nnz;
	if( A->is_splited )
	{
		for(i=0; i<n; i++)
		{
			y[i] = A->D->value[i] * x[i];
		}
		for(k=0; k<A->L->nnz; k++)
		{
			i = A->L->row[k];
			j = A->L->col[k];
			y[i] += A->L->value[k] * x[j];
		}
		for(k=0; k<A->U->nnz; k++)
		{
			i = A->U->row[k];
			j = A->U->col[k];
			y[i] += A->U->value[k] * x[j];
		}
	}
	else
	{
		for(i=0; i<n; i++)
		{
			y[i] = 0.0;
		}
		for(k=0; k<nnz; k++)
		{
			i = A->row[k];
			j = A->col[k];
			y[i] += A->value[k] * x[j];
		}
	}
}

void lis_matvect_coo(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k;
	int n,nnz;

	n    = A->n;
	nnz  = A->nnz;
	if( A->is_splited )
	{
		for(i=0;i<n;i++)
		{
			y[i] = A->D->value[i] * x[i];
		}
		for(k=0;k<A->L->nnz;k++)
		{
			i = A->L->row[k];
			j = A->L->col[k];
			y[j] += A->L->value[k] * x[i];
		}
		for(k=0;k<A->U->nnz;k++)
		{
			i = A->U->row[k];
			j = A->U->col[k];
			y[j] += A->U->value[k] * x[i];
		}
	}
	else
	{
		for(i=0; i<n; i++)
		{
			y[i] = 0.0;
		}
		for(k=0; k<nnz; k++)
		{
			i = A->row[k];
			j = A->col[k];
			y[j] += A->value[k] * x[i];
		}
	}
}
