/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_crs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int			i,j,k,h;
	int			n,np,nl,nz;
	int			is,ie,js,jj;
	int			is0;
	int			ie0;
	int			j0;
	int			*jj0;
	LIS_SCALAR			*vv0;
	LIS_SCALAR			t0;

	n     = A->n;
	if( A->is_splited )
	{
		#pragma omp parallel for private(i,j,is,ie,j0,t0)
		for(i=0;i<n;i++)
		{
			t0  = A->D->value[i]*x[i];

			is = A->L->ptr[i];
			ie = A->L->ptr[i+1];
			for(j=is;j<ie-0;j+=1)
			{
				j0  = A->L->index[j+0];
				t0 += A->L->value[j+ 0]*x[j0];
			}
			is = A->U->ptr[i];
			ie = A->U->ptr[i+1];
			for(j=is;j<ie-0;j+=1)
			{
				j0  = A->U->index[j+0];
				t0 += A->U->value[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
	else
	{
		jj0 = A->index;
		vv0 = A->value;
		#pragma omp parallel for private(i,j,is,ie,j0,t0)
		for(i=0;i<n;i++)
		{
			t0  = 0;

			is = A->ptr[i];
			ie = A->ptr[i+1];
			for(j=is;j<ie-0;j+=1)
			{
				j0 = jj0[j+0];
				t0 += vv0[j+ 0]*x[j0];
			}
			y[i] = t0;
		}
	}
}

void lis_matvect_crs(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je,jj;
	int n,np,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	np   = A->np;
	if( A->is_splited )
	{
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR),"lis_matvect_crs::w" );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<n; i++)
				{
					js = A->L->ptr[i];
					je = A->L->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->L->index[j];
						w[jj] += A->L->value[j] * t;
					}
					js = A->U->ptr[i];
					je = A->U->ptr[i+1];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->U->index[j];
						w[jj] += A->U->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = A->D->value[i] * x[i] + t;
				}
			}
			lis_free(w);
		#else
			for(i=0; i<np; i++)
			{
				y[i] = A->D->value[i] * x[i];
			}
			for(i=0; i<n; i++)
			{
				js = A->L->ptr[i];
				je = A->L->ptr[i+1];
				t = x[i];
				for(j=js;j<je;j++)
				{
					jj  = A->L->index[j];
					y[jj] += A->L->value[j] * t;
				}
			}
			for(i=0; i<n; i++)
			{
				js = A->U->ptr[i];
				je = A->U->ptr[i+1];
				t = x[i];
				for(j=js;j<je;j++)
				{
					jj  = A->U->index[j];
					y[jj] += A->U->value[j] * t;
				}
			}
		#endif
	}
	else
	{
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR),"lis_matvect_crs::w" );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<n; i++)
				{
					js = A->ptr[i];
					je = A->ptr[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
				}
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			for(i=0; i<np; i++)
			{
				y[i] = 0.0;
			}
			for(i=0; i<n; i++)
			{
				js = A->ptr[i];
				je = A->ptr[i+1];
				t = x[i];
				for(j=js;j<je;j++)
				{
					jj  = A->index[j];
					y[jj] += A->value[j] * t;
				}
			}
		#endif
	}
}


