/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_dns(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie;
	int np,n,nprocs,my_rank;
	LIS_SCALAR t;

	n = A->n;
	np = A->np;
	#ifdef _OPENMP
		nprocs  = omp_get_max_threads();
	#else
		nprocs  = 1;
	#endif
	#pragma omp parallel private(i,j,is,ie,my_rank)
	{
		#ifdef _OPENMP
			my_rank = omp_get_thread_num();
		#else
			my_rank = 0;
		#endif
		LIS_GET_ISIE(my_rank,nprocs,n,is,ie);

		for(i=is;i<ie;i++)
		{
			y[i] = 0;
		}
		for(j=0;j<np;j++)
		{
			for(i=is;i<ie;i++)
			{
				y[i] += A->value[j*n+i] * x[j];
			}
		}
	}
}

void lis_matvect_dns(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie;
	int np,n,nprocs,my_rank;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n  = A->n;
	np = A->np;
	#ifdef _OPENMP
		nprocs = omp_get_max_threads();
		w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR),"lis_matvect_dns::w" );
		#pragma omp parallel private(i,j,t,is,ie,my_rank)
		{
			my_rank = omp_get_thread_num();
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
			memset( &w[my_rank*np], 0, np*sizeof(LIS_SCALAR) );
			for(j=0;j<np;j++)
			{
				t = 0.0;
				for(i=is;i<ie;i++)
				{
					t += A->value[j*n+i] * x[i];
				}
				w[my_rank*np + j] = t;
			}
			#pragma omp barrier
			#pragma omp for 
			for(i=0;i<np;i++)
			{
				t = 0.0;
				for(j=0;j<nprocs;j++)
				{
					t += w[j*np+i];
				}
				y[i] = t;
			}
		}
		lis_free(w);
	#else
		for(j=0;j<np;j++)
		{
			t = 0.0;
			for(i=0;i<n;i++)
			{
				t += A->value[j*n+i] * x[i];
			}
			y[j] = t;
		}
	#endif
}
