/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_ell(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,jj,is,ie;
	int n,maxnzr,nprocs,my_rank;

	n      = A->n;
	if( A->is_splited )
	{
		#pragma cdir nodep
		for(i=0; i<n; i++)
		{
			y[i] = A->D->value[i]*x[i];
		}
		for(j=0;j<A->L->maxnzr;j++)
		{
			jj = j*n;
			#pragma cdir nodep
			for(i=0;i<n;i++)
			{
				y[i] += A->L->value[jj + i] * x[A->L->index[jj + i]];
			}
		}
		for(j=0;j<A->U->maxnzr;j++)
		{
			jj = j*n;
			#pragma cdir nodep
			for(i=0;i<n;i++)
			{
				y[i] += A->U->value[jj + i] * x[A->U->index[jj + i]];
			}
		}
	}
	else
	{
		maxnzr = A->maxnzr;
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
		#else
			nprocs = 1;
		#endif
		#pragma omp parallel private(i,j,jj,is,ie,my_rank)
		{
			#ifdef _OPENMP
				my_rank = omp_get_thread_num();
			#else
				my_rank = 0;
			#endif
			LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
			#pragma cdir nodep
			for(i=is;i<ie;i++)
			{
				y[i] = 0.0;
			}
			for(j=0;j<maxnzr;j++)
			{
				jj = j*n;
				#pragma cdir nodep
				for(i=is;i<ie;i++)
				{
					y[i] += A->value[jj + i] * x[A->index[jj + i]];
				}
			}
		}
	}
}

void lis_matvect_ell(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,js,je,jj;
	int n,np,maxnzr,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n      = A->n;
	np     = A->np;
	if( A->is_splited )
	{
		#pragma cdir nodep
		for(i=0; i<n; i++)
		{
			y[i] = A->D->value[i]*x[i];
		}
		for(j=0;j<A->L->maxnzr;j++)
		{
			jj = j*n;
			#pragma cdir nodep
			for(i=0;i<n;i++)
			{
				y[A->L->index[jj + i]] += A->L->value[jj + i] * x[i];
			}
		}
		for(j=0;j<A->U->maxnzr;j++)
		{
			jj = j*n;
			#pragma cdir nodep
			for(i=0;i<n;i++)
			{
				y[A->U->index[jj + i]] += A->U->value[jj + i] * x[i];
			}
		}
	}
	else
	{
		#ifdef _OPENMP
			maxnzr = A->maxnzr;
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR),"lis_matvect_ell::w" );
			#pragma omp parallel private(i,j,t,jj,k,is,ie)
			{
				k = omp_get_thread_num();
				LIS_GET_ISIE(k,nprocs,n,is,ie);
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				for(j=0;j<maxnzr;j++)
				{
					jj = j*n;
					#pragma cdir nodep
					for(i=is;i<ie;i++)
					{
						w[k*np + A->index[jj + i]] += A->value[jj + i] * x[i];
					}
				}
				#pragma omp barrier
				#pragma cdir nodep
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			maxnzr = A->maxnzr;
			#pragma cdir nodep
			for(i=0; i<n; i++)
			{
				y[i] = 0.0;
			}
			for(j=0;j<maxnzr;j++)
			{
				jj = j*n;
				#pragma cdir nodep
				for(i=0;i<n;i++)
				{
					y[A->index[jj + i]] += A->value[jj + i] * x[i];
				}
			}
		#endif
	}
}
