/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

void lis_matvec_msr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,js,je,jj;
	int n;
	LIS_SCALAR t;

	n      = A->n;
	if( A->is_splited )
	{
		n    = A->n;
		#pragma omp parallel for private(i,j,js,je,t,jj)
		for(i=0; i<n; i++)
		{
			t = A->D->value[i] * x[i];
			js = A->L->index[i];
			je = A->L->index[i+1];
			for(j=js;j<je;j++)
			{
				jj = A->L->index[j];
				t += A->L->value[j] * x[jj];
			}
			js = A->U->index[i];
			je = A->U->index[i+1];
			for(j=js;j<je;j++)
			{
				jj = A->U->index[j];
				t += A->U->value[j] * x[jj];
			}
			y[i] = t;
		}
	}
	else
	{
		#pragma omp parallel for private(i,j,js,je,t,jj)
		for(i=0; i<n; i++)
		{
			t = A->value[i] * x[i];
			js = A->index[i];
			je = A->index[i+1];
			for(j=js;j<je;j++)
			{
				jj = A->index[j];
				t += A->value[j] * x[jj];
			}
			y[i] = t;
		}
	}
}

void lis_matvect_msr(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,js,je,jj;
	int n,np,nprocs;
	LIS_SCALAR t;
	LIS_SCALAR *w;

	n    = A->n;
	np   = A->np;

	if( A->is_splited )
	{
		for(i=0; i<n; i++)
		{
			y[i] = A->D->value[i] * x[i];
		}
		for(i=0; i<n; i++)
		{
			t = x[i];
			js = A->L->index[i];
			je = A->L->index[i+1];
			for(j=js;j<je;j++)
			{
				jj = A->L->index[j];
				y[jj] += A->L->value[j] * t;
			}
			js = A->U->index[i];
			je = A->U->index[i+1];
			for(j=js;j<je;j++)
			{
				jj = A->U->index[j];
				y[jj] += A->U->value[j] * t;
			}
		}
	}
	else
	{
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			w = (LIS_SCALAR *)lis_malloc( nprocs*np*sizeof(LIS_SCALAR),"lis_matvect_msr::w" );
			#pragma omp parallel private(i,j,js,je,t,jj,k)
			{
				k = omp_get_thread_num();
				#pragma omp for
				for(j=0;j<nprocs;j++)
				{
					memset( &w[j*np], 0, np*sizeof(LIS_SCALAR) );
				}
				#pragma omp for 
				for(i=0; i<n; i++)
				{
					js = A->index[i];
					je = A->index[i+1];
					t = x[i];
					for(j=js;j<je;j++)
					{
						jj  = k*np+A->index[j];
						w[jj] += A->value[j] * t;
					}
					w[k*np+i] += A->value[i] * x[i];
				}
				#pragma omp for 
				for(i=0;i<np;i++)
				{
					t = 0.0;
					for(j=0;j<nprocs;j++)
					{
						t += w[j*np+i];
					}
					y[i] = t;
				}
			}
			lis_free(w);
		#else
			for(i=0; i<n; i++)
			{
				y[i] = A->value[i] * x[i];
			}
			for(i=0; i<n; i++)
			{
				t = x[i];
				js = A->index[i];
				je = A->index[i+1];
				for(j=js;j<je;j++)
				{
					jj = A->index[j];
					y[jj] += A->value[j] * t;
				}
			}
		#endif
	}
}
