/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef USE_QUAD_PRECISION
#ifdef _WIN32
	#include <float.h>
#endif
#endif
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

#ifdef USE_QUAD_PRECISION
#undef __FUNC__
#define __FUNC__ "lisqp_x87_fpu_init"
void lisqp_x87_fpu_init(unsigned int *cw_old)
{
#ifdef HAS_X87_FPU
#ifdef _WIN32
	unsigned int cw = _control87(0, 0);
	_control87(0x00010000, 0x00030000);
	*cw_old = cw;
	cw = _control87(0, 0);
#else
	int cw,cw_new;
	asm volatile ("fnstcw %0":"=m" (cw));
	cw_new = (cw & ~0x0300) | 0x0200;
	asm volatile ("fldcw %0": :"m" (cw_new));
	*cw_old = cw;
#endif
#else
	*cw_old = 0;
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_x87_fpu_finalize"
void lisqp_x87_fpu_finalize(unsigned int cw)
{
#ifdef HAS_X87_FPU
#ifdef _WIN32
    _control87(cw, 0xFFFFFFFF);
#else
	asm volatile ("fldcw %0": :"m" (cw));
#endif
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_minus"
void  lisqp_minus(LIS_QDOUBLE *a)
{
	a->hi = -a->hi;
	a->lo = -a->lo;
}

#undef __FUNC__
#define __FUNC__ "lisqp_zero"
void  lisqp_zero(LIS_QDOUBLE *a)
{
	a->hi = 0.0;
	a->lo = 0.0;
}

#undef __FUNC__
#define __FUNC__ "lisqp_one"
void  lisqp_one(LIS_QDOUBLE *a)
{
	a->hi = 1.0;
	a->lo = 0.0;
}

#undef __FUNC__
#define __FUNC__ "lisqp_max"
void  lisqp_max(LIS_QDOUBLE *a, LIS_QDOUBLE *b, LIS_QDOUBLE *c)
{
	if( (a->hi > b->hi) || ((a->hi==b->hi) && (a->lo > b->lo)) )
	{
		c->hi = a->hi;
		c->lo = a->lo;
	}
	else
	{
		c->hi = b->hi;
		c->lo = b->lo;
	}
}

#undef __FUNC__
#define __FUNC__ "lisqp_min"
void  lisqp_min(LIS_QDOUBLE *a, LIS_QDOUBLE *b, LIS_QDOUBLE *c)
{
	if( (a->hi < b->hi) || ((a->hi==b->hi) && (a->lo < b->lo)) )
	{
		c->hi = a->hi;
		c->lo = a->lo;
	}
	else
	{
		c->hi = b->hi;
		c->lo = b->lo;
	}
}

#undef __FUNC__
#define __FUNC__ "lisqp_npower"
void lisqp_npower(LIS_QDOUBLE *a, const LIS_QDOUBLE b, const int n)
{
	int			N;
	LIS_QDOUBLE	r,s;

	if( n==0 )
	{
		if( b.hi==0 && b.lo==0 )
		{
			printf("ERROR\n");
			a->hi = 0;
			a->lo = 0;
		}
		else
		{
			a->hi = 1.0;
			a->lo = 0.0;
		}
		return;
	}

	r    = b;
	s.hi = 1.0;
	s.lo = 0.0;
	N    = abs(n);

	if( N>1 )
	{
		while( N>0 )
		{
			if( N%2==1 )
			{
				lisqp_mul(&s,&s,&r);
			}
			N /= 2;
			if( N>0 ) lisqp_sqr(&r,&r);
		}
	}
	else
	{
		s = r;
	}

	if( n<0 )
	{
		lisqp_div_d_dd(a,1.0,&s);
		return;
	}

	*a = s;
}


#undef __FUNC__
#define __FUNC__ "lisqp_add"
void lisqp_add(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c)
{
#ifndef USE_SSE2
	double sh,eh,sl,el,th,tl;

	sh =  b->hi + c->hi;
	th =  sh - b->hi;
	eh =  (b->hi - (sh - th)) + (c->hi - th);
	sl =  b->lo + c->lo;
	tl =  sl - b->lo;
	el =  (b->lo - (sl - tl)) + (c->lo - tl);
	eh += sl;
	th = sh + eh;
	eh = eh - (th - sh);
	eh += el;
	a->hi = th + eh;
	a->lo = eh  - (a->hi - th);
#else
	__m128d xb,xc,xs,xe,xt;

	xb = _mm_loadu_pd((double *)b); /* b->lo,b->hi */
	xc = _mm_loadu_pd((double *)c); /* c->lo,c->hi */
	xs = _mm_add_pd(xb,xc);  /* sl,sh */
	xe = _mm_sub_pd(xs,xb);
	xt = _mm_sub_pd(xs,xe);
	xc = _mm_sub_pd(xc,xe);
	xb = _mm_sub_pd(xb,xt);
	xb = _mm_add_pd(xb,xc);  /* el,eh */
	xe = _mm_unpackhi_pd(xb,xb); /* el */
	xc = _mm_unpackhi_pd(xs,xs); /* sl */
	xt = xs;                     /* sh */
	xb = _mm_add_sd(xb,xc);
	xt = _mm_add_sd(xt,xb);
	xs = _mm_sub_sd(xt,xs);
	xb = _mm_sub_sd(xb,xs);  
	xb = _mm_add_sd(xb,xe);  /* eh */
	xc = _mm_add_sd(xt,xb);  /* a->hi */
	_mm_store_sd(&a->hi,xc);
	xc = _mm_sub_sd(xc,xt);
	xb = _mm_sub_sd(xb,xc);
	_mm_store_sd(&a->lo,xb);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_add_dd_d"
void lisqp_add_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c)
{
#ifndef USE_SSE2
	double s,e,w;

	s = b->hi + c;
	w = s - b->hi;
	e = (b->hi - (s - w)) + (c - w);
	e = e + b->lo;
	a->hi = s + e;
	a->lo = e  - (a->hi - s);
#else
	__m128d xb,xc,xl,xs,xt,xw;

	xb = _mm_load_sd(&b->hi);
	xl = _mm_load_sd(&b->lo);
	xc = _mm_load_sd(&c);
	xs = _mm_add_sd(xb,xc);
	xw = _mm_sub_sd(xs,xb);
	xt = _mm_sub_sd(xs,xw);
	xc = _mm_sub_sd(xc,xw);
	xb = _mm_sub_sd(xb,xt);
	xb = _mm_add_sd(xb,xc);
	xb = _mm_add_sd(xb,xl);
	xt = _mm_add_sd(xs,xb);
	_mm_store_sd(&a->hi,xt);
	xt = _mm_sub_sd(xt,xs);
	xb = _mm_sub_sd(xb,xt);
	_mm_store_sd(&a->lo,xb);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_add_d_d"
void lisqp_add_d_d(LIS_QDOUBLE *a, const double b, const double c)
{
#ifndef USE_SSE2
	double v;

	a->hi =  b + c;
	    v = a->hi - b;
	a->lo =  (b - (a->hi - v)) + (c - v);
#else
	__m128d xb,xc,xr,xt,xw;

	xb = _mm_load_sd(&b);
	xc = _mm_load_sd(&c);
	xr = _mm_add_sd(xb,xc);
	_mm_store_sd(&a->hi,xr);
	xt = _mm_sub_sd(xr,xb);
	xw = _mm_sub_sd(xr,xt);
	xc = _mm_sub_sd(xc,xt);
	xb = _mm_sub_sd(xb,xw);
	xb = _mm_add_sd(xb,xc);
	_mm_store_sd(&a->lo,xb);
#endif
}


#undef __FUNC__
#define __FUNC__ "lisqp_sub"
void lisqp_sub(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c)
{
#ifndef USE_SSE2
	double sh,eh,sl,el,th,tl;

	sh =  b->hi - c->hi;
	th =  sh - b->hi;
	eh =  (b->hi - (sh - th)) - (c->hi + th);
	sl =  b->lo - c->lo;
	tl =  sl - b->lo;
	el =  (b->lo - (sl - tl)) - (c->lo + tl);
	eh += sl;
	th = sh + eh;
	eh = eh - (th - sh);
	eh += el;
	a->hi = th + eh;
	a->lo = eh  - (a->hi - th);
#else
	__m128d xb,xc,xs,xe,xt;

	xb = _mm_loadu_pd((double *)b); /* b->lo,b->hi */
	xc = _mm_loadu_pd((double *)c); /* c->lo,c->hi */
	xs = _mm_sub_pd(xb,xc);  /* sl,sh */
	xe = _mm_sub_pd(xs,xb);
	xt = _mm_sub_pd(xs,xe);
	xc = _mm_add_pd(xc,xe);
	xb = _mm_sub_pd(xb,xt);
	xb = _mm_sub_pd(xb,xc);  /* el,eh */
	xe = _mm_unpackhi_pd(xb,xb); /* el */
	xc = _mm_unpackhi_pd(xs,xs); /* sl */
	xt = xs;                     /* sh */
	xb = _mm_add_sd(xb,xc);
	xt = _mm_add_sd(xt,xb);
	xs = _mm_sub_sd(xt,xs);
	xb = _mm_sub_sd(xb,xs);  
	xb = _mm_add_sd(xb,xe);  /* eh */
	xc = _mm_add_sd(xt,xb);  /* a->hi */
	_mm_store_sd(&a->hi,xc);
	xc = _mm_sub_sd(xc,xt);
	xb = _mm_sub_sd(xb,xc);
	_mm_store_sd(&a->lo,xb);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_sub_dd_d"
void lisqp_sub_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c)
{
#ifndef USE_SSE2
	double s,e,w;

	s = b->hi - c;
	w = s - b->hi;
	e = (b->hi - (s - w)) - (c + w);
	e = e + b->lo;
	a->hi = s + e;
	a->lo = e  - (a->hi - s);
#else
	__m128d xb,xc,xl,xs,xt,xw;

	xb = _mm_load_sd(&b->hi);
	xl = _mm_load_sd(&b->lo);
	xc = _mm_load_sd(&c);
	xs = _mm_sub_sd(xb,xc);
	xw = _mm_sub_sd(xs,xb);
	xt = _mm_sub_sd(xs,xw);
	xc = _mm_add_sd(xc,xw);
	xb = _mm_sub_sd(xb,xt);
	xb = _mm_sub_sd(xb,xc);
	xb = _mm_add_sd(xb,xl);
	xt = _mm_add_sd(xs,xb);
	_mm_store_sd(&a->hi,xt);
	xt = _mm_sub_sd(xt,xs);
	xb = _mm_sub_sd(xb,xt);
	_mm_store_sd(&a->lo,xb);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_mul"
void lisqp_mul(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c)
{
#ifndef USE_SSE2
#ifndef HAS_FMA
	double p1, p2;
	double t,bhi,blo,chi,clo;

	p1 =  b->hi * c->hi;
	t   = SPLITTER * b->hi;
	bhi = t - (t-b->hi);
	blo = b->hi - bhi;
	t   = SPLITTER * c->hi;
	chi = t - (t-c->hi);
	clo = c->hi - chi;
	p2 = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo;
	p2 += (b->hi * c->lo);
	p2 += (b->lo * c->hi);
	a->hi = p1 + p2;
	a->lo = p2  - (a->hi - p1);
#else
	double p1, p2;

	p1 = -b->hi * c->hi;
	p2 = b->hi * c->hi + p1;
	p2 += (b->hi * c->lo);
	p2 += (b->lo * c->hi);
	a->hi = p2 - p1;
	a->lo = p2  - (a->hi + p1);
#endif
#else
	__m128d bch,bcl,xs,xt,xr,xw,xe,xb;

	xs  = _mm_set_pd(SPLITTER,SPLITTER);
	bch = _mm_set_pd(b->hi,c->hi);
	xb  = _mm_unpackhi_pd(bch,bch); /* bh,bh */
	bcl = _mm_set_pd(c->lo,b->lo);
	xb  = _mm_mul_sd(xb,bch);
	xs  = _mm_mul_pd(xs,bch);
	xt  = _mm_sub_pd(xs,bch);
	xs  = _mm_sub_pd(xs,xt);
	xt  = _mm_mul_pd(bch,bcl);
	bch = _mm_sub_pd(bch,xs);
	xe  = _mm_unpackhi_pd(xs,bch);  /* blo,sh */
	xw  = _mm_unpacklo_pd(xs,bch);  /* clo,sl */
	bch = _mm_unpackhi_pd(bch,xw);  /* clo,blo */
	xw  = _mm_mul_pd(xw,xe);
	xs  = _mm_mul_pd(xs,bch);
	bch = _mm_unpackhi_pd(xs,xs);
	bcl = _mm_unpackhi_pd(xw,xw);
	xe  = _mm_unpackhi_pd(xt,xt);
	xw  = _mm_sub_sd(xw,xb);
	xw  = _mm_add_sd(xw,bch);
	xw  = _mm_add_sd(xw,xs);
	xw  = _mm_add_sd(xw,bcl);
	xw  = _mm_add_sd(xw,xe);
	xw  = _mm_add_sd(xw,xt);
	xt  = _mm_add_sd(xw,xb);
	_mm_store_sd(&a->hi,xt);
	xt  = _mm_sub_sd(xt,xb);
	xw  = _mm_sub_sd(xw,xt);
	_mm_store_sd(&a->lo,xw);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_mul_dd_d"
 void lisqp_mul_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c)
{
#ifndef USE_SSE2
#ifndef HAS_FMA
  double p1, p2,ahi,alo,bhi,blo,t;

	p1  = b->hi * c;
	t   = SPLITTER * b->hi;
	ahi = t - (t-b->hi);
	alo = b->hi - ahi;
	t   = SPLITTER * c;
	bhi = t - (t-c);
	blo = c - bhi;
	p2  = ((ahi*bhi-p1)+ahi*blo+alo*bhi)+alo*blo;
	p2 += (b->lo * c);
	a->hi = p1 + p2;
	a->lo = p2  - (a->hi - p1);
#else
	double p1, p2;

	p1 = -b->hi * c;
	p2 = b->hi * c + p1;
	p2 += (b->lo * c);
	a->hi = p2 - p1;
	a->lo = p2  - (a->hi + p1);
#endif
#else
	__m128d bch,bcl,xs,xt,xr,xw,xe,xb;

	xs  = _mm_set_pd(SPLITTER,SPLITTER);
	bch = _mm_set_pd(b->hi,c);
	xb  = _mm_unpackhi_pd(bch,bch); /* bh,bh */
	bcl = _mm_load_sd(&b->lo);
	xb  = _mm_mul_sd(xb,bch);
	bcl = _mm_mul_sd(bcl,bch);
	xs  = _mm_mul_pd(xs,bch);
	xt  = _mm_sub_pd(xs,bch);
	xs  = _mm_sub_pd(xs,xt);
	bch = _mm_sub_pd(bch,xs);
	xe  = _mm_unpackhi_pd(xs,bch);  /* blo,sh */
	xw  = _mm_unpacklo_pd(xs,bch);  /* clo,sl */
	bch = _mm_unpackhi_pd(bch,xw);  /* clo,blo */
	xw  = _mm_mul_pd(xw,xe);
	xs  = _mm_mul_pd(xs,bch);
	bch = _mm_unpackhi_pd(xs,xs);
	xt  = _mm_unpackhi_pd(xw,xw);
	xw  = _mm_sub_sd(xw,xb);
	xw  = _mm_add_sd(xw,bch);
	xw  = _mm_add_sd(xw,xs);
	xw  = _mm_add_sd(xw,xt);
	xw  = _mm_add_sd(xw,bcl);
	xt  = _mm_add_sd(xw,xb);
	_mm_store_sd(&a->hi,xt);
	xt  = _mm_sub_sd(xt,xb);
	xw  = _mm_sub_sd(xw,xt);
	_mm_store_sd(&a->lo,xw);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_sqr"
void lisqp_sqr(LIS_QDOUBLE *a, const LIS_QDOUBLE *b)
{
#ifndef USE_SSE2
#ifndef HAS_FMA
	double p1, p2,hi,lo,t;

	p1 =  b->hi * b->hi;
	t   = SPLITTER * b->hi;
	hi = t - (t-b->hi);
	lo = b->hi - hi;
	p2 = ((hi*hi-p1)+2.0*hi*lo)+lo*lo;
	p2 += (2.0*b->hi * b->lo);
	p2 += (b->lo * b->lo);
	a->hi = p1 + p2;
	a->lo = p2  - (a->hi - p1);
#else
	double p1, p2;

	p1 = -b->hi * b->hi;
	p2 = b->hi * b->hi + p1;
	p2 += (2.0*b->hi * b->lo);
	p2 += (b->lo * b->lo);
	a->hi = p2 - p1;
	a->lo = p2  - (a->hi + p1);
#endif
#else
	__m128d bh,bl,p1,hi,lo,t0;

	hi  = _mm_set_sd(SPLITTER);
	bh  = _mm_load_sd(&b->hi);
	bl  = _mm_load_sd(&b->lo);
	p1  = _mm_mul_sd(bh,bh); /* p1=b->hi*b->hi    */
	hi  = _mm_mul_sd(hi,bh); /* hi=SPLITTER*b->hi */
	t0  = _mm_sub_sd(hi,bh); /* t0=hi-b->hi */
	hi  = _mm_sub_sd(hi,t0); /* hi=hi-t0 */
	lo  = _mm_sub_sd(bh,hi); /* lo=b->hi-hi */
	t0  = _mm_mul_sd(hi,hi); /* t0=hi*hi */
	hi  = _mm_add_sd(hi,hi); /* hi=hi+hi */
	hi  = _mm_mul_sd(hi,lo); /* hi=hi*lo */
	lo  = _mm_mul_sd(lo,lo); /* lo=lo*lo */
	t0  = _mm_sub_sd(t0,p1); /* t0=t0-p1 */
	t0  = _mm_add_sd(t0,hi); /* t0=t0+hi */
	t0  = _mm_add_sd(t0,lo); /* t0=t0+lo */
	bh  = _mm_add_sd(bh,bh); /* bh=bh+bh */
	bh  = _mm_mul_sd(bh,bl); /* bh=bh*b->lo */
	bl  = _mm_mul_sd(bl,bl); /* bl=bl*bl */
	t0  = _mm_add_sd(t0,bh); /* t0=t0+bh */
	t0  = _mm_add_sd(t0,bl); /* t0=t0+bl */
	hi  = _mm_add_sd(p1,t0); /* hi=p1+t0 */
	_mm_store_sd(&a->hi,hi);
	hi  = _mm_sub_sd(hi,p1); /* hi=hi-p1 */
	t0  = _mm_sub_sd(t0,hi); /* t0=t0-hi */
	_mm_store_sd(&a->lo,t0);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_div"
void lisqp_div(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const LIS_QDOUBLE *c)
{
#ifndef USE_SSE2
#ifndef HAS_FMA
	double q1, q2;
	double s1, s2;
	double t;
	double p1,p2,chi,clo,bhi,blo;
	LIS_QDOUBLE r;

	q1  = b->hi / c->hi;
	p1 =  c->hi * q1;
	t   = SPLITTER * c->hi;
	bhi = t - (t-c->hi);
	blo = c->hi - bhi;
	t   = SPLITTER * q1;
	chi = t - (t-q1);
	clo = q1 - chi;
	p2 = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo;

	p2 += (c->lo * q1);
	r.hi = p1 + p2;
	r.lo = p2  - (r.hi - p1);
	s1 =  b->hi - r.hi;
	 t =  s1 - b->hi;
	s2 =  (b->hi - (s1 - t)) - (r.hi + t);
	s2 -= r.lo;
	s2 += b->lo;
	q2  = (s1+s2) / c->hi;
	a->hi = q1 + q2;
	a->lo = q2  - (a->hi - q1);
#else
	double q1, q2;
	double s1, s2;
	double t;
	double p1,p2;
	LIS_QDOUBLE r;

	q1   = b->hi / c->hi;

	p1   = -c->hi * q1;
	p2   = c->hi * q1 + p1;
	p2  += (c->lo * q1);
	r.hi = p2 - p1;
	r.lo = p2  - (r.hi + p1);

	s1 =  b->hi - r.hi;
	 t =  s1 - b->hi;
	s2 =  (b->hi - (s1 - t)) - (r.hi + t);
	s2 -= r.lo;
	s2 += b->lo;
	q2  = (s1+s2) / c->hi;
	a->hi = q1 + q2;
	a->lo = q2  - (a->hi - q1);
#endif
#else
	__m128d xb,xc,xs,xe,xw,xq,xr,xt;

	xs  = _mm_set_pd(SPLITTER,SPLITTER);
	xb  = _mm_set_pd(b->lo,b->hi);
	xc  = _mm_set_pd(c->lo,c->hi);
	xq  = xb;                      /* bl,bh */
	xw  = xc;                      /* cl,ch */
	xq  = _mm_div_sd(xq,xc);       /* bl,q1 */
	xw  = _mm_unpacklo_pd(xw,xq);  /* q1,ch */
	xc  = _mm_move_sd(xc,xq);      /* cl,q1 */
	xq  = xw;                      /* q1,ch */
	xs  = _mm_mul_pd(xs,xw);
	xt  = _mm_sub_pd(xs,xw);
	xs  = _mm_sub_pd(xs,xt);
	xc  = _mm_mul_pd(xc,xw);
	xw  = _mm_sub_pd(xw,xs);
	xt  = xs;
	xr  = xw;
	xt  = _mm_unpackhi_pd(xt,xt);  /* sh,sh */
	xr  = _mm_unpackhi_pd(xr,xr);  /* wh,wh */
	xe  = xt;
	xt  = _mm_mul_sd(xt,xw);
	xe  = _mm_mul_sd(xe,xs);
	xw  = _mm_mul_sd(xw,xr);
	xs  = _mm_mul_sd(xs,xr);
	xr  = xc;                      /* th,tl */
	xr  = _mm_unpackhi_pd(xr,xr);  /* th,th */
	xe  = _mm_sub_sd(xe,xc);
	xe  = _mm_add_sd(xe,xs);
	xe  = _mm_add_sd(xe,xt);
	xe  = _mm_add_sd(xe,xw);
	xe  = _mm_add_sd(xe,xr);       /* ??,p2 */
	xr  = xc;
	xr  = _mm_add_sd(xr,xe);  /* rh=tl+p2 */
	xw  = xr;
	xw  = _mm_sub_sd(xw,xc);  /* t5=rh-tl */
	xs  = xb;
	xs  = _mm_sub_sd(xs,xr);  /* s1=bh-rh */
	xe  = _mm_sub_sd(xe,xw);  /* p2=p2-t5 */
	xt  = xs;
	xt  = _mm_sub_sd(xt,xb);  /* t4=s1-bh */
	xr  = _mm_add_sd(xr,xt);  /* rh=rh+t4 */
	xw  = xs;
	xw  = _mm_sub_sd(xw,xt);  /* t4=s1-t4 */
	xt  = xb;
	xt  = _mm_unpackhi_pd(xt,xt);  /* bl,bl */
	xb  = _mm_sub_sd(xb,xw);
	xb  = _mm_sub_sd(xb,xr);
	xb  = _mm_sub_sd(xb,xe);
	xb  = _mm_add_sd(xb,xt);       /* ??,s2 */
	xt  = xq;
	xt  = _mm_unpackhi_pd(xt,xt);  /* q1,q1 */
	xe  = xt;
	xs  = _mm_add_sd(xs,xb);
	xs  = _mm_div_sd(xs,xq);
	xt  = _mm_add_sd(xt,xs);
	_mm_store_sd(&a->hi,xt);
	xt  = _mm_sub_sd(xt,xe);
	xs  = _mm_sub_sd(xs,xt);
	_mm_store_sd(&a->lo,xs);
#endif
}

#undef __FUNC__
#define __FUNC__ "lisqp_div_dd_d"
void lisqp_div_dd_d(LIS_QDOUBLE *a, const LIS_QDOUBLE *b, const double c)
{
	double p1, p2;
	double q1, q2;
	double s,e,t,v;
	double ahi,alo,bhi,blo;

	q1  = b->hi / c;
	p1  =  q1 * c;
	t   = SPLITTER * q1;
	ahi = t - (t-q1);
	alo = q1 - ahi;
	t   = SPLITTER * c;
	bhi = t - (t-c);
	blo = c - bhi;
	p2  = ((ahi*bhi-p1)+ahi*blo+alo*bhi)+alo*blo;
	s   = b->hi - p1;
	v   = s - b->hi;
	e   = (b->hi - (s - v)) - (p1 + v);
	e  += b->lo;
	e  -= p2;
	q2  = (s+e) / c;
	a->hi = q1 + q2;
	a->lo = q2  - (a->hi - q1);
}

#undef __FUNC__
#define __FUNC__ "lisqp_div_d_dd"
void lisqp_div_d_dd(LIS_QDOUBLE *a, const double b, const LIS_QDOUBLE *c)
{
	LIS_QDOUBLE t;

	t.hi = b;
	t.lo = 0.0;
	lisqp_div(a,&t,c);
}

#undef __FUNC__
#define __FUNC__ "lisqp_sqrt"
void lisqp_sqrt(LIS_QDOUBLE *a, const LIS_QDOUBLE *b)
{
#ifndef USE_SSE2
	double x,hi,lo,t;
	double sh,eh,sl,el,th,tl;
	LIS_QDOUBLE xx,sx;

	if( b->hi==0 )
	{
		lisqp_zero(a);
		return;
	}
	if( b->hi<0 )
	{
		printf("ERROR bh=%e\n",b->hi);
		lisqp_zero(a);
		return;
	}

	x     = 1.0 / sqrt(b->hi);
	xx.hi = b->hi * x;
	x     = x * 0.5;

	sx.hi =  xx.hi * xx.hi;
	t   = SPLITTER * xx.hi;
	hi = t - (t-xx.hi);
	lo = xx.hi - hi;
	sx.lo = ((hi*hi-sx.hi)+2.0*hi*lo)+lo*lo;

	sh =  b->hi - sx.hi;
	th =  sh - b->hi;
	eh =  (b->hi - (sh - th)) - (sx.hi + th);
	sl =  b->lo - sx.lo;
	tl =  sl - b->lo;
	el =  (b->lo - (sl - tl)) - (sx.lo + tl);
	eh += sl;
	th = sh + eh;
	eh = eh - (th - sh);
	eh += el;
	a->hi = th + eh;
	x  = a->hi * x;
	a->hi = xx.hi + x;
	a->lo = x  - (a->hi - xx.hi);
#else
	__m128d hi,lo,xs,xt,xr,xw,t0,t1,t2;
	__m128d xb,bh,p1,xe,xc;

	if( b->hi==0 )
	{
		lisqp_zero(a);
		return;
	}
	if( b->hi<0 )
	{
		printf("ERROR bh=%e\n",b->hi);
		lisqp_zero(a);
		return;
	}

	hi  = _mm_set_sd(SPLITTER);
	xb = _mm_loadu_pd((const double *)b); /*b->lo,b->hi*/
	bh = xb;
	t0 = _mm_set_sd(1.0);
	t1 = _mm_set_sd(0.5);
	t2 = _mm_sqrt_pd(bh);
	t0 = _mm_div_sd(t0,t2);
	t2 = _mm_mul_sd(bh,t0); /* xx */
	t0 = _mm_mul_sd(t0,t1); /* x */
	p1  = _mm_mul_sd(t2,t2); /* sx.hi */
	hi  = _mm_mul_sd(hi,t2);
	t1  = _mm_sub_sd(hi,t2);
	hi  = _mm_sub_sd(hi,t1);
	lo  = _mm_sub_sd(t2,hi);
	t1  = _mm_mul_sd(hi,hi); /* hi*hi */
	hi  = _mm_add_sd(hi,hi);
	hi  = _mm_mul_sd(hi,lo); /* 2.0*hi*lo */
	lo  = _mm_mul_sd(lo,lo);
	t1  = _mm_sub_sd(t1,p1);
	t1  = _mm_add_sd(t1,hi);
	t1  = _mm_add_sd(t1,lo); /* sx.lo */
	p1 = _mm_unpacklo_pd(p1,t1); /* sx.lo,sx.hi */
	xs = _mm_sub_pd(xb,p1);  /* sl,sh */
	xe = _mm_sub_pd(xs,xb);
	xt = _mm_sub_pd(xs,xe);
	p1 = _mm_add_pd(p1,xe);
	xb = _mm_sub_pd(xb,xt);
	xb = _mm_sub_pd(xb,p1);  /* el,eh */
	xe = _mm_unpackhi_pd(xb,xb); /* el */
	xc = _mm_unpackhi_pd(xs,xs); /* sl */
	xt = xs;                     /* sh */
	xb = _mm_add_sd(xb,xc);
	xt = _mm_add_sd(xt,xb);
	xs = _mm_sub_sd(xt,xs);
	xb = _mm_sub_sd(xb,xs);
	xb = _mm_add_sd(xb,xe);
	xt = _mm_add_sd(xt,xb);  /* a->hi */
	t0 = _mm_mul_sd(t0,xt);
	xr = _mm_add_sd(t2,t0);
	_mm_store_sd(&a->hi,xr);
	xt = _mm_sub_sd(xr,t2);
	t0 = _mm_sub_sd(t0,xt);
	_mm_store_sd(&a->lo,t0);
#endif
}
#endif

