/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/
#undef __FUNC__
#define __FUNC__ "lis_precon_create_adds"
int lis_precon_create_adds(LIS_SOLVER solver, LIS_PRECON precon)
{
	int		i,j;
	int		precon_type,worklen;
	int		err;
	LIS_VECTOR *work;

	LIS_DEBUG_FUNC_IN;

	precon_type = solver->options[LIS_OPTIONS_PRECON];
	worklen     = 2;
	work        = (LIS_VECTOR *)lis_malloc( worklen*sizeof(LIS_VECTOR),"lis_precon_create_adds::work" );
	if( work==NULL )
	{
		LIS_SETERR_MEM(worklen*sizeof(LIS_VECTOR));
		return LIS_OUT_OF_MEMORY;
	}
	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		for(i=0;i<worklen;i++)
		{
			err = lis_vector_duplicate(solver->A,&work[i]);
			if( err ) break;
		}
	}
	else
	{
		for(i=0;i<worklen;i++)
		{
			err = lis_vector_duplicateex(LIS_PRECISION_QUAD,solver->A,&work[i]);
			if( err ) break;
		}
	}
	if( i<worklen )
	{
		for(j=0;j<i;j++) lis_vector_destroy(work[j]);
		lis_free(work);
		return err;
	}
	precon->worklen = worklen;
	precon->work    = work;

	err = lis_precon_create_xxx[precon_type](solver,precon);
	if( err )
	{
		lis_precon_destroy(precon);
		return err;
	}

    precon->A       = solver->A;
    precon->is_copy = LIS_FALSE;

	LIS_DEBUG_FUNC_OUT;
	return err;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_adds"
int lis_psolve_adds(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,k,n,np,iter,ptype;
	LIS_SCALAR *b,*x,*w,*r;
	LIS_VECTOR W,R;
	LIS_PRECON	precon;


	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	n     = precon->A->n;
	np    = precon->A->np;
	W     = precon->work[0];
	R     = precon->work[1];
	b     = B->value;
	x     = X->value;
	w     = W->value;
	r     = R->value;
	iter  = solver->options[LIS_OPTIONS_ADDS_ITER];
	ptype = solver->options[LIS_OPTIONS_PRECON];

	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		lis_vector_set_all(0.0,X);
		lis_vector_copy(B,R);
		for(k=0;k<iter+1;k++)
		{
			for(i=n;i<np;i++)
			{
				r[i] = 0.0;
			}

			lis_psolve_xxx[ptype](solver,R,W);
			for(i=0;i<n;i++)
			{
				x[i] += w[i];
			}
		
			if(k==iter) break;

			lis_matvec(precon->A,X,R);
			for(i=0;i<n;i++)
			{
				r[i] = b[i] - r[i];
			}
		}
	}
	else
	{
		lis_vector_set_all(0.0,X);
		lis_vector_copy(B,R);
		for(k=0;k<iter+1;k++)
		{
			for(i=n;i<np;i++)
			{
				r[i] = 0.0;
			}

			lis_psolve_xxx[ptype](solver,R,W);
			for(i=0;i<n;i++)
			{
				x[i] += w[i];
			}
		
			if(k==iter) break;

			X->precision = LIS_PRECISION_DEFAULT;
			lis_matvec(precon->A,X,R);
			X->precision = LIS_PRECISION_QUAD;
			for(i=0;i<n;i++)
			{
				r[i] = b[i] - r[i];
			}
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_adds"
int lis_psolvet_adds(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,k,n,np,iter,ptype;
	LIS_SCALAR *b,*x,*w,*r;
	LIS_VECTOR W,R;
	LIS_PRECON	precon;


	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	n     = precon->A->n;
	np    = precon->A->np;
	W     = precon->work[0];
	R     = precon->work[1];
	b     = B->value;
	x     = X->value;
	w     = W->value;
	r     = R->value;
	iter  = solver->options[LIS_OPTIONS_ADDS_ITER];
	ptype = solver->options[LIS_OPTIONS_PRECON];

	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		lis_vector_set_all(0.0,X);
		lis_vector_copy(B,R);
		for(k=0;k<iter+1;k++)
		{
			for(i=n;i<np;i++)
			{
				r[i] = 0.0;
			}

			lis_psolvet_xxx[ptype](solver,R,W);
			for(i=0;i<n;i++)
			{
				x[i] += w[i];
			}
		
			if(k==iter) break;

			lis_matvect(precon->A,X,R);
			for(i=0;i<n;i++)
			{
				r[i] = b[i] - r[i];
			}
		}
	}
	else
	{
		lis_vector_set_all(0.0,X);
		lis_vector_copy(B,R);
		for(k=0;k<iter+1;k++)
		{
			for(i=n;i<np;i++)
			{
				r[i] = 0.0;
			}

			lis_psolvet_xxx[ptype](solver,R,W);
			for(i=0;i<n;i++)
			{
				x[i] += w[i];
			}
		
			if(k==iter) break;

			X->precision = LIS_PRECISION_DEFAULT;
			lis_matvect(precon->A,X,R);
			X->precision = LIS_PRECISION_QUAD;
			for(i=0;i<n;i++)
			{
				r[i] = b[i] - r[i];
			}
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

