/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_precon_create_iluc"
int lis_precon_create_iluc(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;
	LIS_MATRIX		A,B;

	LIS_DEBUG_FUNC_IN;

	switch( solver->A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_matrix_split(solver->A);
		if( err )
		{
			return err;
		}
		err = lis_precon_create_iluc_crs(solver,precon);
		break;
	default:
		A = solver->A;
		err = lis_matrix_duplicate(A,&B);
		if( err ) return err;
		lis_matrix_set_type(B,LIS_MATRIX_CRS);
		err = lis_matrix_convert(A,B);
		if( err ) return err;
		solver->A = B;
		err = lis_matrix_split(solver->A);
		if( err )
		{
			return err;
		}
		err = lis_precon_create_iluc_crs(solver,precon);
		lis_matrix_destroy(B);
		solver->A = A;
		break;
	}

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_precon_create_iluc_crs"
int lis_precon_create_iluc_crs(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;
	int				i,j,k,l,ii,jj,kk,ll;
	int				n,nnz,annz,lfil,len;
	int				cz,cw;
	int				*iw,*wc,*wl,*iz,*zc,*zl;
	int				*index,*ptr;
	LIS_SCALAR		gamma,val,t,dd,tol,toldd;
	LIS_SCALAR		*z,*w,*tmp;
/*	LIS_SCALAR		*ljj; */
	LIS_SCALAR		*value;
	LIS_SCALAR		m;
	LIS_MATRIX		A,L,U;
	LIS_VECTOR		D;

	LIS_DEBUG_FUNC_IN;


	A      = solver->A;
	n      = A->n;
	tol    = solver->params[LIS_PARAMS_DROP-LIS_OPTIONS_LEN];
	m      = solver->params[LIS_PARAMS_RATE-LIS_OPTIONS_LEN];
	gamma  = solver->params[LIS_PARAMS_GAMMA-LIS_OPTIONS_LEN];
	annz   = 10+A->nnz / A->n;
	lfil   = (int)((double)A->nnz/(2.0*n))*m;

	L      = NULL;
	U      = NULL;


	err = lis_matrix_create_rco(n,0,A->comm,0,NULL,&L);
	if( err )
	{
		return err;
	}
	err = lis_matrix_create_rco(n,0,A->comm,0,NULL,&U);
	if( err )
	{
		return err;
	}
	err = lis_vector_duplicate(A,&D);
	if( err )
	{
		return err;
	}

	tmp   = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR),"lis_precon_create_iluc_crs::tmp");
	if( tmp==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(LIS_SCALAR));
		return LIS_OUT_OF_MEMORY;
	}
	w   = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR),"lis_precon_create_iluc_crs::w");
	if( w==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(LIS_SCALAR));
		return LIS_OUT_OF_MEMORY;
	}
	z   = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR),"lis_precon_create_iluc_crs::z");
	if( z==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(LIS_SCALAR));
		return LIS_OUT_OF_MEMORY;
	}
	/*
	ljj   = (LIS_SCALAR *)lis_malloc(n*sizeof(LIS_SCALAR),"lis_precon_create_iluc_crs::ljj");
	if( ljj==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(LIS_SCALAR));
		return LIS_OUT_OF_MEMORY;
	}
	*/

	iw   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_iluc_crs::iw");
	if( iw==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		return LIS_OUT_OF_MEMORY;
	}
	wc   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_iluc_crs::wc");
	if( wc==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		return LIS_OUT_OF_MEMORY;
	}
	wl   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_iluc_crs::wl");
	if( wl==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		return LIS_OUT_OF_MEMORY;
	}
	iz   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_iluc_crs::iz");
	if( iw==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		return LIS_OUT_OF_MEMORY;
	}
	zc   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_iluc_crs::zc");
	if( wc==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		return LIS_OUT_OF_MEMORY;
	}
	zl   = (int *)lis_malloc(n*sizeof(int),"lis_precon_create_iluc_crs::zl");
	if( wl==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		return LIS_OUT_OF_MEMORY;
	}

	lis_matrix_sort_crs(A);
	nnz = A->L->nnz;
	err = lis_matrix_malloc_crs(n,nnz,&ptr,&index,&value);

	for(i=0;i<n;i++) iw[i] = 0;
/*	for(i=0;i<n;i++) ljj[i] = 1.0;*/
	for(i=0;i<n;i++)
	{
		for(j=A->L->ptr[i];j<A->L->ptr[i+1];j++)
		{
			jj = A->L->index[j];
			iw[jj]++;
		}
	}
	ptr[0] = 0;
	for(i=0;i<n;i++)
	{
		ptr[i+1] = ptr[i] + iw[i];
		iw[i]    = ptr[i];
	}
	for(i=0;i<n;i++)
	{
		for(j=A->L->ptr[i];j<A->L->ptr[i+1];j++)
		{
			k        = A->L->index[j];
			l        = iw[k];
			value[l] = A->L->value[j];
			index[l] = i;
			iw[k]++;
		}
	}


	for(i=0;i<n;i++)
	{
		D->value[i] = gamma * A->D->value[i];
		zl[i]       = -1;
		wl[i]       = -1;
		zc[i]       = 0;
		wc[i]       = 0;
	}


	for(k=0;k<n;k++)
	{
		cz = 0;
		cw = 0;
		/* z_j = a_kj (j>=k) */
		for(j=A->U->ptr[k];j<A->U->ptr[k+1];j++)
		{
			jj       = A->U->index[j];
#ifdef USE_MPI
			if( jj>n-1 ) continue;
#endif
			z[jj]    = A->U->value[j];
			iz[cz++] = jj;
			zc[jj]   = 1;
		}

		/* w_j = a_jk (j>k) */
		for(j=ptr[k];j<ptr[k+1];j++)
		{
			jj       = index[j];
			w[jj]    = value[j];
			iw[cw++] = jj;
			wc[jj]   = 1;
		}

		/* z_j = z_j - l_ki*u_ij */
		i = wl[k];
		while( i>=0 )
		{
			ii  = wc[i];
			kk  = zc[i];
			val = L->w_value[i][ii];
			for(j=kk;j<U->w_row[i];j++)
			{
				jj     = U->w_index[i][j];
				if( jj==k ) continue;
				if( zc[jj]==1 )
				{
					z[jj]   -= val*U->w_value[i][j];
				}
				else
				{
					z[jj]    = -val*U->w_value[i][j];
					iz[cz++] = jj;
					zc[jj]   = 1;
				}
			}
			ii++;
			wc[i] = ii;
			if( ii<L->w_row[i] )
			{
				kk     = L->w_index[i][ii];
				ll     = i;
				i      = wl[ll];
				wl[ll] = wl[kk];
				wl[kk] = ll;
			}
			else
			{
				i      = wl[i];
			}
		}

		/* w_j = w_j - l_ji*u_ik */
		i = zl[k];
		while( i>=0 )
		{
			ii  = zc[i];
			kk  = wc[i];
			val = U->w_value[i][ii];
			for(j=kk;j<L->w_row[i];j++)
			{
				jj     = L->w_index[i][j];
				if( wc[jj]==1 )
				{
					w[jj]   -= val*L->w_value[i][j];
				}
				else
				{
					w[jj]    = -val*L->w_value[i][j];
					iw[cw++] = jj;
					wc[jj]   = 1;
				}
			}
			ii++;
			zc[i] = ii;
			if( ii<U->w_row[i] )
			{
				kk     = U->w_index[i][ii];
				ll     = i;
				i      = zl[ll];
				zl[ll] = zl[kk];
				zl[kk] = ll;
			}
			else
			{
				i      = zl[i];
			}
		}

		toldd       = fabs(D->value[k])*tol;
		D->value[k] = 1.0/D->value[k];
		t           = D->value[k];
		if( cz<cw )
		{
			for(j=0;j<cz;j++)
			{
				jj = iz[j];
				if( wc[jj] )
				{
					D->value[jj] -= z[jj]*w[jj]*t;
				}
			}
		}
		else
		{
			for(j=0;j<cw;j++)
			{
				jj = iw[j];
				if( zc[jj] )
				{
					D->value[jj] -= z[jj]*w[jj]*t;
				}
			}
		}
		/* drop U */
		nnz = 0;
		for(j=0;j<cz;j++)
		{
			jj = iz[j];
			if( fabs(z[jj])>toldd )
			{
				iz[nnz++] = jj;
			}
			else
			{
				z[k] = z[k] + fabs(z[jj]);
				zc[jj] = 0;
			}
		}
		len = _min(lfil,nnz);
		for(j=0;j<nnz;j++) tmp[j] = fabs(z[j]);
		lis_sort_di(0,nnz-1,tmp,iz);
		lis_sort_i(0,len-1,iz);
		U->w_row[k] = len;
		if( len>0 )
		{
			U->w_index[k] = (int *)lis_malloc(len*sizeof(int),"lis_precon_create_iluc_crs::U->w_index");
			U->w_value[k] = (LIS_SCALAR *)lis_malloc(len*sizeof(LIS_SCALAR),"lis_precon_create_iluc_crs::U->w_value");
		}
		for(j=0;j<len;j++)
		{
			jj = iz[j];
			U->w_index[k][j] = jj;
			U->w_value[k][j] = z[jj];
		}
		for(j=len;j<cz;j++) zc[iz[j]] = 0;
		cz = nnz;

		/* drop L */
		nnz = 0;
		for(j=0;j<cw;j++)
		{
			jj = iw[j];
			if( fabs(w[jj])>toldd )
			{
				iw[nnz++] = jj;
			}
			else
			{
/*				ljj[jj] += fabs(w[jj]);*/
				wc[jj] = 0;
			}
		}
		len = _min(lfil,nnz);
		for(j=0;j<nnz;j++) tmp[j] = fabs(w[j]);
		lis_sort_di(0,nnz-1,tmp,iw);
		lis_sort_i(0,len-1,iw);
		L->w_row[k] = len;
		if( len>0 )
		{
			L->w_index[k] = (int *)lis_malloc(len*sizeof(int),"lis_precon_create_iluc_crs::L->w_index");
			L->w_value[k] = (LIS_SCALAR *)lis_malloc(len*sizeof(LIS_SCALAR),"lis_precon_create_iluc_crs::L->w_value");
		}
		for(j=0;j<len;j++)
		{
			jj = iw[j];
			L->w_index[k][j] = jj;
			L->w_value[k][j] = w[jj]*t;
		}
		for(j=len;j<cw;j++) wc[iw[j]] = 0;
		cw = nnz;

		/**/
		for(j=0;j<cz;j++) zc[iz[j]] = 0;
		for(j=0;j<cw;j++) wc[iw[j]] = 0;
		if(U->w_row[k]>0)
		{
			jj     = iz[0];
			zc[k]  = 0;
			zl[k]  = zl[jj];
			zl[jj] = k;
		}
		if(L->w_row[k]>0)
		{
			jj     = iw[0];
			wc[k]  = 0;
			wl[k]  = wl[jj];
			wl[jj] = k;
		}
	}

/*	for(i=0;i<n;i++) ljj[i] = 1.0 / ljj[i];
	L->work    = ljj;*/
	precon->L  = L;
	precon->U  = U;
	precon->D  = D;

	lis_free2(12,tmp,w,z,iw,wc,wl,iz,zc,zl,ptr,index,value);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_psolve_iluc"
int lis_psolve_iluc(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	LIS_DEBUG_FUNC_IN;

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */
/*
	n       = precon->Pmat->n;
	b       = B->value;
	x       = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		t = b[i];
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t;
	}
	for(i=ie-1; i>=is; i--)
	{
		t = x[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t * precon->Pmat->value[i];
	}
}
*/
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;

#else
	int i,j,jj,n;
	LIS_SCALAR w;
	LIS_SCALAR *b,*x;
	LIS_MATRIX L,U;
	LIS_VECTOR D;
	LIS_PRECON	precon;
	LIS_QUAD_DECLAR;
	#ifdef USE_QUAD_PRECISION
		LIS_QDOUBLE w1,w2;
		LIS_SCALAR *xl;
	#endif

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	L = precon->L;
	U = precon->U;
	D = precon->D;
	n = L->n;
	b = B->value;
	x = X->value;
	#ifdef USE_QUAD_PRECISION
		xl = X->value_lo;
	#endif

	#ifdef USE_QUAD_PRECISION
		if( B->precision==LIS_PRECISION_DEFAULT )
		{
	#endif
			lis_vector_copy(B,X);
			for(i=0; i<n; i++)
			{
				for(j=0;j<L->w_row[i];j++)
				{
					jj     = L->w_index[i][j];
					x[jj] -= L->w_value[i][j] * x[i];
				}
/*				x[i] = x[i] * L->work[i];*/
			}
			for(i=n-1; i>=0; i--)
			{
				w = x[i];
				for(j=0;j<U->w_row[i];j++)
				{
					jj = U->w_index[i][j];
					w -= U->w_value[i][j] * x[jj];
				}
				x[i] = w * D->value[i];
			}
	#ifdef USE_QUAD_PRECISION
		}
		else
		{
			lis_vector_copyex_mm(B,X);
			for(i=0; i<n; i++)
			{
				for(j=0;j<L->w_row[i];j++)
				{
					jj     = L->w_index[i][j];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(x[jj],xl[jj],x[jj],xl[jj],x[i],xl[i],-L->w_value[i][j]);
					#else
						LIS_QUAD_FMAD_SSE2(x[jj],xl[jj],x[jj],xl[jj],x[i],xl[i],-L->w_value[i][j]);
					#endif
					/* x[jj] -= L->w_value[i][j] * x[i]; */
				}
/*				#ifndef USE_SSE2
					LIS_QUAD_MULD(x[i],xl[i],x[i],xl[i],L->work[i]);
				#else
					LIS_QUAD_MULD_SSE2(x[i],xl[i],x[i],xl[i],L->work[i]);
				#endif*/
			}
			for(i=n-1; i>=0; i--)
			{
				w1.hi = x[i];
				w1.lo = xl[i];
				/* t = x[i]; */
				for(j=0;j<U->w_row[i];j++)
				{
					jj = U->w_index[i][j];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(w1.hi,w1.lo,w1.hi,w1.lo,x[jj],xl[jj],-U->w_value[i][j]);
					#else
						LIS_QUAD_FMAD_SSE2(w1.hi,w1.lo,w1.hi,w1.lo,x[jj],xl[jj],-U->w_value[i][j]);
					#endif
					/* t -= U->w_value[i][j] * x[jj]; */
				}
				#ifndef USE_SSE2
					LIS_QUAD_MULD(x[i],xl[i],w1.hi,w1.lo,D->value[i]);
				#else
					LIS_QUAD_MULD_SSE2(x[i],xl[i],w1.hi,w1.lo,D->value[i]);
				#endif
				/* x[i] = t * D->value[i]; */
			}
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_iluc"
int lis_psolvet_iluc(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	LIS_DEBUG_FUNC_IN;

	/*
	 *  (LU)'x = b
	 *  LU     = (D + L*A) * (I + D^-1 * U*A)
	 */
/*
	n       = precon->Pmat->n;
	b       = B->value;
	x       = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		x[i] = b[i];
	}
	for(i=is; i<ie; i++)
	{
		x[i] = x[i] * precon->Pmat->value[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) x[jj] -= precon->Pmat->value[j] * x[i];
		}
	}
	for(i=ie-1; i>=is; i--)
	{
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) x[jj] -= precon->Pmat->value[j] * x[i];
		}
	}
}
*/
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	int i,j,jj,n;
	LIS_SCALAR w;
	LIS_SCALAR *b,*x;
	LIS_MATRIX L,U;
	LIS_VECTOR D;
	LIS_PRECON	precon;
	LIS_QUAD_DECLAR;
	#ifdef USE_QUAD_PRECISION
		LIS_QDOUBLE w1,w2;
		LIS_SCALAR *xl;
	#endif

	/*
	 *  (LU)'x = b
	 *  U'x=b
	 *  L'x=b
	 */


	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	L = precon->L;
	U = precon->U;
	D = precon->D;
	n = L->n;
	b = B->value;
	x = X->value;
	#ifdef USE_QUAD_PRECISION
		xl = X->value_lo;
	#endif

	#ifdef USE_QUAD_PRECISION
		if( B->precision==LIS_PRECISION_DEFAULT )
		{
	#endif
			lis_vector_copy(B,X);
			for(i=0; i<n; i++)
			{
				x[i] = x[i] * D->value[i];
				for(j=0;j<U->w_row[i];j++)
				{
					jj     = U->w_index[i][j];
					x[jj] -= U->w_value[i][j] * x[i];
				}
			}
			for(i=n-2; i>=0; i--)
			{
				w = x[i];
				for(j=0;j<L->w_row[i];j++)
				{
					jj  = L->w_index[i][j];
					w  -= L->w_value[i][j] * x[jj];
				}
				x[i] = w;
			}
	#ifdef USE_QUAD_PRECISION
		}
		else
		{
			lis_vector_copyex_mm(B,X);
			for(i=0; i<n; i++)
			{
				#ifndef USE_SSE2
					LIS_QUAD_MULD(x[i],xl[i],x[i],xl[i],D->value[i]);
				#else
					LIS_QUAD_MULD_SSE2(x[i],xl[i],x[i],xl[i],D->value[i]);
				#endif
				/* x[i] = x[i] * D->value[i]; */
				for(j=0;j<U->w_row[i];j++)
				{
					jj     = U->w_index[i][j];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(x[jj],xl[jj],x[jj],xl[jj],x[i],xl[i],-U->w_value[i][j]);
					#else
						LIS_QUAD_FMAD_SSE2(x[jj],xl[jj],x[jj],xl[jj],x[i],xl[i],-U->w_value[i][j]);
					#endif
					/* x[jj] -= U->w_value[i][j] * x[i]; */
				}
			}
			for(i=n-2; i>=0; i--)
			{
				w1.hi = x[i];
				w1.lo = xl[i];
				/* t = x[i]; */
				for(j=0;j<L->w_row[i];j++)
				{
					jj  = L->w_index[i][j];
					#ifndef USE_SSE2
						LIS_QUAD_FMAD(w1.hi,w1.lo,w1.hi,w1.lo,x[jj],xl[jj],-L->w_value[i][j]);
					#else
						LIS_QUAD_FMAD_SSE2(w1.hi,w1.lo,w1.hi,w1.lo,x[jj],xl[jj],-L->w_value[i][j]);
					#endif
					/* t  -= L->w_value[i][j] * x[jj]; */
				}
				x[i]  = w1.hi;
				xl[i] = w1.lo;
				/* x[i] = t; */
			}
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

