/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_precon_create_jacobi"
int lis_precon_create_jacobi(LIS_SOLVER solver, LIS_PRECON precon)
{
	int			i,n,err;
	LIS_SCALAR	*D;

	LIS_DEBUG_FUNC_IN;

	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		err = lis_vector_duplicate(solver->A, &precon->D);
	}
	else
	{
		err = lis_vector_duplicateex(LIS_PRECISION_QUAD,solver->A, &precon->D);
	}
	if( err )
	{
		return err;
	}

	lis_matrix_get_diagonal(solver->A, precon->D);
	lis_vector_reciprocal(precon->D);

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_jacobi"
int lis_psolve_jacobi(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,n;
	LIS_SCALAR *b,*x,*d;
	LIS_PRECON	precon;
	LIS_QUAD_DECLAR;
	#ifdef USE_QUAD_PRECISION
		LIS_SCALAR *xl;
	#endif

	LIS_DEBUG_FUNC_IN;

	/*
	 *  Mx = b
	 *  M  = D
	 */

	precon = solver->precon;
	n = precon->D->n;
	d = precon->D->value;
	b = B->value;
	x = X->value;
	#ifdef USE_QUAD_PRECISION
		xl = X->value_lo;
	#endif

	#ifdef USE_QUAD_PRECISION
		if( B->precision==LIS_PRECISION_DEFAULT )
		{
	#endif
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				x[i] = b[i] * d[i];
			}
	#ifdef USE_QUAD_PRECISION
		}
		else
		{
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				#ifndef USE_SSE2
					LIS_QUAD_MULD(x[i],xl[i],B->value[i],B->value_lo[i],d[i]);
				#else
					LIS_QUAD_MULD_SSE2(x[i],xl[i],B->value[i],B->value_lo[i],d[i]);
				#endif
				/* x[i] = b[i] * d[i]; */
			}
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_jacobi"
int lis_psolvet_jacobi(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,n;
	LIS_SCALAR *b,*x,*d;
	LIS_PRECON	precon;
	LIS_QUAD_DECLAR;
	#ifdef USE_QUAD_PRECISION
		LIS_SCALAR *xl;
	#endif

	LIS_DEBUG_FUNC_IN;

	/*
	 *  Mx = b
	 *  M  = D
	 */

	precon = solver->precon;
	n = precon->D->n;
	d = precon->D->value;
	b = B->value;
	x = X->value;
	#ifdef USE_QUAD_PRECISION
		xl = X->value_lo;
	#endif

	#ifdef USE_QUAD_PRECISION
		if( B->precision==LIS_PRECISION_DEFAULT )
		{
	#endif
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				x[i] = b[i] * d[i];
			}
	#ifdef USE_QUAD_PRECISION
		}
		else
		{
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				#ifndef USE_SSE2
					LIS_QUAD_MULD(x[i],xl[i],B->value[i],B->value_lo[i],d[i]);
				#else
					LIS_QUAD_MULD_SSE2(x[i],xl[i],B->value[i],B->value_lo[i],d[i]);
				#endif
				/* x[i] = b[i] * d[i]; */
			}
		}
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
