/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/
#undef __FUNC__
#define __FUNC__ "lis_precon_create_ssor"
int lis_precon_create_ssor(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;
	LIS_SCALAR		w;
	LIS_MATRIX		A;

	LIS_DEBUG_FUNC_IN;

	A   = solver->A;
	w   = solver->params[LIS_PARAMS_SSOR_W-LIS_OPTIONS_LEN];

	err = lis_matrix_split(A);
	if( err )
	{
		return err;
	}
	if( A->use_wd!=LIS_SOLVER_SOR )
	{
		if( !A->WD )
		{
			err = lis_matrix_diag_duplicate(A->D,&A->WD);
			if( err ) return err;
		}
		lis_matrix_diag_copy(A->D,A->WD);
		lis_matrix_diag_scale(w,A->WD);
		lis_matrix_diag_inverse(A->WD);
		A->use_wd = LIS_SOLVER_SOR;
	}

	precon->A       = A;
	precon->is_copy = LIS_FALSE;

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_ssor"
int lis_psolve_ssor(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i;
	LIS_MATRIX		A;
	LIS_PRECON	precon;

	/*
	 *  Mx = b
	 *  M  = (D/w + L) * (I + w*D^-1 * U)
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	A = precon->A;

	lis_matrix_solve(A,B,X,LIS_MATRIX_SSOR);


	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_ssor"
int lis_psolvet_ssor(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i;
	LIS_MATRIX		A;
	LIS_PRECON	precon;

	/*
	 *  M'x = b
	 *  M'  = (I + U' * w*D^-1) * (D/w + L')
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	A = precon->A;

	lis_matrix_solvet(A,B,X,LIS_MATRIX_SSOR);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
