/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_input
 ************************************************/
int lis_input_option(LIS_MATRIX_INPUT_OPTION *option, char *path);

#undef __FUNC__
#define __FUNC__ "lis_input"
int lis_input(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, char *filename)
{
	int		err;
	int		fileformat;
	char	buf[256],banner[128];
	FILE	*file;
	LIS_MATRIX_INPUT_OPTION	option;

	LIS_DEBUG_FUNC_IN;

	err = lis_matrix_check(A,LIS_MATRIX_CHECK_NULL);
	if( err ) return err;
	if( b!=NULL && x!=NULL )
	{
		err = lis_vector_check(b,LIS_VECTOR_CHECK_NULL);
		if( err ) return err;
		err = lis_vector_check(x,LIS_VECTOR_CHECK_NULL);
		if( err ) return err;
	}

	if( filename==NULL )
	{
		LIS_SETERR(LIS_ERR_ILL_ARG,"filname is NULL\n");
		return LIS_ERR_ILL_ARG;
	}
	file = fopen(filename, "r");
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n",filename);
		return LIS_ERR_FILE_IO;
	}

	/* file format check */
	if( fgets(buf, 256, file) == NULL )
	{
		fclose(file);
		return LIS_ERR_FILE_IO;
	}
	sscanf(buf, "%s", banner);
	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER)) == 0)
	{
		fileformat = LIS_FMT_MM;
	}
/*	else if( strncmp(banner, LISBanner, strlen(LISBanner)) == 0)
	{
		fileformat = LIS_FMT_LIS;
	}
	else if( strncmp(banner, ITBLBanner, strlen(ITBLBanner)) == 0)
	{
		fileformat = LIS_FMT_ITBL;
	}
*/
	else
	{
		fileformat = LIS_FMT_HB;
	}
	rewind(file);

/*
	if( fileformat==LIS_FMT_FREE )
	{
		fclose(file);
		err = lis_input_option(&option, filename);
		if( err ) return err;
		file = fopen(option.filename, "r");
		if( file==NULL )
		{
			LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n",filename);
			return LIS_ERR_FILE_IO;
		}
	}
*/

	switch( fileformat )
	{
	case LIS_FMT_MM:
		err = lis_input_mm(A,b,x,file);
		break;
	case LIS_FMT_HB:
		err = lis_input_hb(A,b,x,file);
		break;
/*
	case LIS_FMT_ITBL:
		err = lis_input_mmm(A,b,x,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	case LIS_FMT_LIS:
		err = lis_input_lis(A,b,x,filename,file,comm,matrix_type,bnr,bnc,row,col);
		break;
	case LIS_FMT_FREE:
		err = lis_input_free(A,b,x,option,file,comm,matrix_type,bnr,bnc,row,col);
		break;
*/
	default:
		fclose(file);
		return err;
	}
	fclose(file);
#ifdef USE_MPI
	MPI_Barrier(A->comm);
#endif

	LIS_DEBUG_FUNC_OUT;
	return err;
}


#undef __FUNC__
#define __FUNC__ "lis_input_option"
int lis_input_option(LIS_MATRIX_INPUT_OPTION *option, char *path)
{
	int i;
	int aorder=0;
	int porder=0;
	FILE *file;
	char str[1024];
	char seps[] = " \t\n\r";
	char *p,*s;

	LIS_DEBUG_FUNC_IN;
/*
	option->fileformat        = LIS_FMT_LIS;
	option->is_solution       = LIS_FALSE;
	option->is_rhs            = LIS_FALSE;
	option->array_len         = 0;
	option->param_len         = 0;
	option->is_array_zero_cut = LIS_FALSE;
	option->is_matrix_split   = LIS_FALSE;
	file = fopen(path, "r");
	if( file==NULL )
	{
		printf("cannot open %s\n", path);
		return LIS_FAILS;
	}
	while( fgets(str,sizeof(str),file)!=NULL )
	{
		p = str;
		while( *p!='\0' ) *p++ = (char)tolower(*p);
		p   = strtok( str, seps );
//		while( *p==' ' || *p==0x09 ) p++;
		if( p==NULL ) continue;
		if( strcmp(p,"filename")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: filename is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			strcpy(option->filename,p);
		}
		else if( strcmp(p,"matrixformat")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: matrixformat is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			for(i=LIS_MATRIX_CRS;i<LIS_MATRIX_DNS+1;i++)
			{
				if( strcmp(p,mattypename[i])==0 ) break;
			}
			if( i==LIS_MATRIX_DNS+1 )
			{
				printf("Error: matrixformat '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
			option->matrix_type_in = i;
		}
		else if( aorder < 9 && strcmp(p,"arrayorder")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: arrayorder is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			for(i=0;i<18;i++)
			{
				if( strcmp(p,arrayorder_name[i])==0 ) break;
			}
			if( i==18 )
			{
				printf("Error: arrayorder '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
			else if( i==7 )
			{
				option->is_solution = LIS_TRUE;
			}
			else if( i==8 )
			{
				option->is_rhs = LIS_TRUE;
			}
			option->array_order[aorder++] = i;
			option->array_len = aorder;
		}
		else if( porder < 9 && strcmp(p,"paramorder")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: paramorder is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			for(i=0;i<12;i++)
			{
				if( strcmp(p,paramorder_name[i])==0 ) break;
			}
			if( i==12 )
			{
				printf("Error: arrayorder '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
			option->param_order[porder++] = i;
			option->param_len = porder;
		}
		else if( strcmp(p,"matrixsplit")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: matrixsplit is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"true")==0 )
			{
				option->is_matrix_split = LIS_TRUE;
			}
			else if( strcmp(p,"false")==0 )
			{
				option->is_matrix_split = LIS_FALSE;
			}
			else
			{
				printf("Error: matrixsplit '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
		else if( strcmp(p,"array0cut")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: array0cut is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"true")==0 )
			{
				option->is_array_zero_cut = LIS_TRUE;
			}
			else if( strcmp(p,"false")==0 )
			{
				option->is_array_zero_cut = LIS_FALSE;
			}
			else
			{
				printf("Error: array0cut '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
		else if( strcmp(p,"arraybase")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: arraybase is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"0")==0 )
			{
				option->array_base_ptr = LIS_MATRIX_ARRAY_BASE_0;
			}
			else if( strcmp(p,"1")==0 )
			{
				option->array_base_ptr = LIS_MATRIX_ARRAY_BASE_1;
			}
			else
			{
				printf("Error: arraybase '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
		else if( strcmp(p,"matrixbase")==0 )
		{
			p = strtok(NULL,seps);
			if( p==NULL )
			{
				printf("Error: matrixbase is undefined\n");
				fclose(file);
				return LIS_FAILS;
			}
			if( strcmp(p,"0")==0 )
			{
				option->array_base_index = LIS_MATRIX_ARRAY_BASE_0;
			}
			else if( strcmp(p,"1")==0 )
			{
				option->array_base_index = LIS_MATRIX_ARRAY_BASE_1;
			}
			else
			{
				printf("Error: matrixbase '%s' is unknown\n",p);
				fclose(file);
				return LIS_FAILS;
			}
		}
	}
	fclose(file);
*/
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_input_vector"
int lis_input_vector(LIS_VECTOR v, char *filename, LIS_Comm comm)
{
	int		fileformat;
	char	buf[256],banner[128];
	int		err;
	FILE	*file;

	if( filename==NULL )
	{
		LIS_SETERR(LIS_ERR_ILL_ARG,"filname is NULL\n");
		return LIS_ERR_ILL_ARG;
	}
	file = fopen(filename, "r");
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n",filename);
		return LIS_ERR_FILE_IO;
	}

	if( fgets(buf, 256, file) == NULL )
	{
		fclose(file);
		return LIS_ERR_FILE_IO;
	}
	sscanf(buf, "%s", banner);
	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER)) == 0)
	{
		fileformat = LIS_FMT_MM;
	}
	else if( strncmp(banner, LISBanner, strlen(LISBanner)) == 0)
	{
		fileformat = LIS_FMT_LIS;
	}
	else
	{
		fileformat = LIS_FMT_PLAIN;
	}
	rewind(file);

	switch( fileformat )
	{
	case LIS_FMT_MM:
		err = lis_input_vector_mm(v,file,comm);
		break;
	case LIS_FMT_LIS:
		err = lis_input_vector_lis(v,filename,file,comm);
		break;
	case LIS_FMT_PLAIN:
		err = lis_input_vector_plain(v,file,comm);
		break;
	}
	fclose(file);
#ifdef USE_MPI
	MPI_Barrier(comm);
#endif
	return err;
}


#undef __FUNC__
#define __FUNC__ "lis_input_vector_mm"
int lis_input_vector_mm(LIS_VECTOR v, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			banner[64], mtx[64], fmt[64], dtype[64], dstruct[64];
	char			*p;
	int				i,k,j;
	int				err;
	int				n,is,ie;
	int				idx;
	LIS_SCALAR		val;


	/* check banner */
	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		LIS_SETERR_FIO;
		return LIS_ERR_FILE_IO;
	}
	sscanf(buf, "%s %s %s %s %s", banner, mtx, fmt, dtype, dstruct);

	for(p=mtx;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=fmt;*p!='\0';p++)     *p = (char)tolower(*p);
	for(p=dtype;*p!='\0';p++)   *p = (char)tolower(*p);
	for(p=dstruct;*p!='\0';p++) *p = (char)tolower(*p);

	if( strncmp(banner, MM_BANNER, strlen(MM_BANNER))!=0 || strncmp(mtx, MM_VEC, strlen(MM_VEC))!=0 )
	{
		LIS_SETERR(LIS_ERR_FILE_IO,"Not Matrix Market banner\n");
		return LIS_ERR_FILE_IO;
	}
	if( strncmp(fmt, MM_FMT, strlen(MM_FMT))!=0 )
	{
		LIS_SETERR(LIS_ERR_FILE_IO,"Not Coodinate format\n");
		return LIS_ERR_FILE_IO;
	}
	if( strncmp(dtype, MM_TYPE_REAL, strlen(MM_TYPE_REAL))!=0 )
	{
		LIS_SETERR(LIS_ERR_FILE_IO,"Not real\n");
		return LIS_ERR_FILE_IO;
	}
	if( strncmp(dstruct, MM_TYPE_GENERAL, strlen(MM_TYPE_GENERAL))!=0 )
	{
		LIS_SETERR(LIS_ERR_FILE_IO,"Not general\n");
		return LIS_ERR_FILE_IO;
	}

	/* check size */		
	do
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			LIS_SETERR_FIO;
			return LIS_ERR_FILE_IO;
		}
	}while( buf[0]=='%' );
	if( sscanf(buf, "%d", &n) != 1 )
	{
		LIS_SETERR_FIO;
		return LIS_ERR_FILE_IO;
	}

	/* read data */
	err = lis_vector_set_size(v,0,n);
	if( err )
	{
		return err;
	}
	lis_vector_get_range(v,&is,&ie);

	for(i=0;i<n;i++)
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			LIS_SETERR_FIO;
			return LIS_ERR_FILE_IO;
		}
		if( sscanf(buf, "%d %lg", &idx, &val) != 2 )
		{
			LIS_SETERR_FIO;
			return LIS_ERR_FILE_IO;
		}
		idx--;
		if( idx>=is && idx<ie )
		{
			v->value[idx-is] = val;
		}
	}
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_input_vector_plain"
int lis_input_vector_plain(LIS_VECTOR v, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE];
	char			*p;
	int				i,k,j;
	int				err;
	int				n,is,ie;
	int				idx;
	LIS_SCALAR		val;


	/* check size */
	n = 0;
	do
	{
		err = fscanf(file, "%lg", &val);
		if( err==1 ) n++;
	}while( err==1 );
	rewind(file);

	/* read data */
	err = lis_vector_set_size(v,0,n);
	if( err )
	{
		return err;
	}
	lis_vector_get_range(v,&is,&ie);

	for(i=0;i<n;i++)
	{
		if( fgets(buf, BUFSIZE, file) == NULL )
		{
			LIS_SETERR_FIO;
			return LIS_ERR_FILE_IO;
		}
		if( i>=is && i<ie )
		{
			if( sscanf(buf, "%lg", &val) != 1 )
			{
				LIS_SETERR_FIO;
				return LIS_ERR_FILE_IO;
			}
			v->value[i-is] = val;
		}
	}
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_input_vector_lis"
int lis_input_vector_lis(LIS_VECTOR v, char *filename, FILE *file, LIS_Comm comm)
{
	char			buf[BUFSIZE],banner[128],mode[128],mattype[128];
	int				in_mode;
	int				err;

	if( fgets(buf, BUFSIZE, file) == NULL )
	{
		LIS_SETERR_FIO;
		return LIS_ERR_FILE_IO;
	}
	buf[10] = '\0';
	sscanf(buf, "%s %s %s", banner, mode, mattype);
	if( strncmp(banner, LISBanner, strlen(LISBanner)) != 0)
	{
		LIS_SETERR(LIS_ERR_FILE_IO,"not lis file format\n");
		return LIS_ERR_FILE_IO;
	}

	in_mode = LIS_FMT_LIS_ASCII;
	if( mode[0]=='B' || mode[0]=='L' )
	{
		fclose(file);
		file = fopen(filename, "rb");
		if( file==NULL )
		{
			LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", filename);
			return LIS_ERR_FILE_IO;
		}
		fread(buf, sizeof(char), 10, file);
		in_mode = 1;
		in_mode = *(char *)&in_mode;
		if( (in_mode==LIS_BINARY_BIG && mode[0]=='L') || (in_mode==LIS_BINARY_LITTLE && mode[0]=='B') )
		{
			in_mode = LIS_TRUE;
		}
		else
		{
			in_mode = LIS_FALSE;
		}
	}

	if( strncmp(mattype, "vec", 3) == 0 )
	{
		if( in_mode==LIS_FMT_LIS_ASCII )
		{
			lis_input_vector_lis_ascii(v,file,comm);
		}
		else
		{
			LIS_SETERR_IMP;
			return LIS_ERR_NOT_IMPLEMENTED;
		}
	}
	else
	{
		LIS_SETERR(LIS_ERR_FILE_IO,"not lis file format\n");
		return LIS_ERR_FILE_IO;
	}

	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_input_vector_lis_ascii"
int lis_input_vector_lis_ascii(LIS_VECTOR v, FILE *file, LIS_Comm comm)
{
	int				n,nprocs,my_rank,pe;
	int				i;
	int				err;
	int				ibuf[10];
	char			cbuf[BUFSIZE];
	char			c;

#ifdef USE_MPI
	MPI_Comm_size(comm,&nprocs);
	MPI_Comm_rank(comm,&my_rank);
#else
	nprocs  = 1;
	my_rank = 0;
#endif

	if( fgets(cbuf, BUFSIZE, file) == NULL )
	{
		LIS_SETERR_FIO;
		return LIS_ERR_FILE_IO;
	}
	if( sscanf(cbuf, "%d",&ibuf[0])!=1 )
	{
		LIS_SETERR_FIO;
		return LIS_ERR_FILE_IO;
	}
	if( nprocs!=ibuf[0] )
	{
		LIS_SETERR2(LIS_ERR_FILE_IO,"The number of PE=(%d) is different (in file PE=%d)\n",nprocs,ibuf[0]);
		return LIS_ERR_FILE_IO;
	}
	pe=-1;
	do
	{
		if( fgets(cbuf, BUFSIZE, file) == NULL )
		{
			LIS_SETERR_FIO;
			return LIS_ERR_FILE_IO;
		}
		if( cbuf[0]=='#' )
		{
			if( sscanf(cbuf, "%c %d %d",&c, &pe, &ibuf[1])!=3 )
			{
				LIS_SETERR_FIO;
				return LIS_ERR_FILE_IO;
			}
		}
	}while(pe!=my_rank);

	n   = ibuf[1];
	err = lis_vector_set_size(v,0,n);
	if( err )
	{
		return err;
	}

	err = lis_fscan_double(n,file,v->value);
	if( err )
	{
		LIS_SETERR_FIO;
		return LIS_ERR_FILE_IO;
	}

	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_fscan_double"
int lis_fscan_double(int n, FILE *file, double val[])
{
	int i,k;

	i=0;
	while( i<n )
	{
		fscanf(file, "%lg", &val[i++]);
	}
	return LIS_SUCCESS;
}

