/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif
#undef malloc

#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include "lislib.h"

#if 1
typedef struct _malloc_address {
	struct _malloc_address *next, *prev;
	void *address, *real_address;
	size_t size;
	#ifdef _DEBUG
		char *tag;
	#endif
} malloc_address;

malloc_address malloc_address_top = 
  {&malloc_address_top, &malloc_address_top, NULL, NULL};
#else
#define MALLOC_ADDRESS_SIZE	1021

typedef struct _malloc_address {
  struct _malloc_address *next, *prev;
  void *address, *real_address;
  size_t size;
} malloc_address;

malloc_address malloc_address_top = 
  {&malloc_address_top, &malloc_address_top, NULL, NULL};

malloc_address *malloc_address_table;
#endif

void *malloc ();
 
void *rpl_malloc( size_t n )
{
	if( n == 0 )
	n = 1;
	return malloc( n );
}

void *lis_malloc( size_t size, char *tag )
{
	union {
		void *ptr;
		unsigned long l;
	} addr;
	malloc_address *ma;
	int sz;

	addr.ptr = malloc(size + 16);
    ma = malloc(sizeof(malloc_address));
    ma->next = &malloc_address_top;
	ma->prev = malloc_address_top.prev;
	ma->prev->next = ma;
	ma->next->prev = ma;
	ma->real_address = addr.ptr;
	ma->size = size;
	addr.l += 0xf;
	addr.l &= ~0xf;
	ma->address = addr.ptr;

	#ifdef _DEBUG
		sz = strlen(tag);
		ma->tag = (char *)malloc(sz+1);
		strcpy(ma->tag,tag);
	#endif
 
  return addr.ptr;
}

void *lis_calloc( size_t size, char *tag )
{
  union {
    void *ptr;
    unsigned long l;
  } addr;
  malloc_address *ma;
  int sz;

	addr.ptr = malloc(size + 16);
    ma = malloc(sizeof(malloc_address));
    ma->next = &malloc_address_top;
	ma->prev = malloc_address_top.prev;
	ma->prev->next = ma;
	ma->next->prev = ma;
	ma->real_address = addr.ptr;
	ma->size = size;
	addr.l += 0xf;
	addr.l &= ~0xf;
	ma->address = addr.ptr;
	memset(addr.ptr,0,size);

	#ifdef _DEBUG
		sz = strlen(tag);
		ma->tag = (char *)malloc(sz+1);
		strcpy(ma->tag,tag);
	#endif

  return addr.ptr;
}

void *lis_realloc( void *p, size_t size )
{
	union {
		void *ptr;
		unsigned long l;
	} addr;
	malloc_address *ma;
	size_t old_size;
	void *real_address,*address;

	ma = malloc_address_top.next;
	while (ma->address)
	{
	    if (ma->address == p) break;
		ma = ma->next;
	}
	if (ma->address)
	{
		old_size       = ma->size;
		addr.ptr       = malloc(size + 16);
		real_address   = addr.ptr;
		addr.l        += 0xf;
		addr.l        &= ~0xf;
		address        = addr.ptr;

		memcpy(address,p,old_size);
		free(ma->real_address);

		ma->address      = address;
		ma->real_address = real_address;
		ma->size         = size;
		return address;
	}
	address = realloc(p,size);
	return address;
}

void lis_free(void *p)
{
  malloc_address *ma;

  ma = malloc_address_top.next;
  while (ma->address) {
    if (ma->address == p) break;
	ma = ma->next;
  }
  if (ma->address) {
    ma->next->prev = ma->prev;
	ma->prev->next = ma->next;
    free(ma->real_address);
	#ifdef _DEBUG
		free(ma->tag);
	#endif
	free(ma);
	return;
  }
  free(p);
  return;
}

void lis_free_mat(LIS_MATRIX A)
{
	malloc_address *ma;
	malloc_address *ma_bak;
	int i;

	ma = malloc_address_top.next;
	for(i=0;i<A->n;i++)
	{
		if( A->w_index[i]==NULL ) continue;
		while (ma->address) {
			if (ma->address == A->w_index[i]) break;
			ma = ma->next;
		}
		if (ma->address) {
			ma->next->prev = ma->prev;
			ma->prev->next = ma->next;
			ma_bak         = ma->next;
			free(ma->real_address);
			#ifdef _DEBUG
				free(ma->tag);
			#endif
			free(ma);
			ma = ma_bak;
		}
		while (ma->address) {
			if (ma->address == A->w_value[i]) break;
			ma = ma->next;
		}
		if (ma->address) {
			ma->next->prev = ma->prev;
			ma->prev->next = ma->next;
			ma_bak         = ma->next;
			free(ma->real_address);
			#ifdef _DEBUG
				free(ma->tag);
			#endif
			free(ma);
			ma = ma_bak;
		}
	}
	return;
}

void lis_free2(int n, ...)
{
	va_list vvlist;
	int		i;
	char	*p;

	va_start( vvlist, n );
	for(i=0;i<n;i++)
	{
		p = va_arg(vvlist, char *);
		if( p )
		{
			lis_free(p);
		}
	}
	va_end(vvlist);
}

void lis_free_all()
{
  malloc_address *ma;
  malloc_address *ma_bak;

  ma = malloc_address_top.next;
  while (ma->address) {
  #ifdef _DEBUG
/*	  lis_printf(LIS_COMM_WORLD,"memory leak: address = %x size=%d\n",ma->address,ma->size);*/
	  lis_printf(LIS_COMM_WORLD,"memory leak: address = %x size=%d (%s)\n",ma->address,ma->size,ma->tag);
  #endif
    ma_bak = ma->next;
	free(ma->real_address);
	#ifdef _DEBUG
		free(ma->tag);
	#endif
	free(ma);
	ma = ma_bak;
  }
  malloc_address_top.next         = &malloc_address_top;
  malloc_address_top.prev         = &malloc_address_top;
  malloc_address_top.address      = NULL;
  malloc_address_top.real_address = NULL;
  malloc_address_top.size         = 0;
  return;
}

int lis_is_malloc( void *p )
{
	malloc_address *ma;

	ma = malloc_address_top.next;
	while (ma->address)
	{
	    if (ma->address == p) break;
		ma = ma->next;
	}
	if (ma->address)
	{
		return LIS_TRUE;
	}
	return LIS_FALSE;
}
