/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_output_mm
 ************************************************/
#undef __FUNC__
#define __FUNC__ "lis_output_mm_vec"
#ifdef USE_MPI
int lis_output_mm_vec(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	FILE			*file;
	LIS_MM_VECFMT	fmt;

	LIS_DEBUG_FUNC_IN;


	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	is        = A->is;

	if( !lis_vector_is_null(b) )
	{
		err = 0;
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				if( format==LIS_FMT_MM )
				{
					file = fopen(path, "a");
				}
				else
				{
					file = fopen(path, "ab");
				}
				if( file==NULL )
				{
					LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
					err = 1;
				}
				else
				{
					if( format==LIS_FMT_MM )
					{
						for(i=0;i<n;i++)
						{
							fprintf(file, "%d %28.20e\n", is+i+1,b->value[i]);
						}
					}
					else
					{
						for(i=0;i<n;i++)
						{
							fmt.i     = is+i+1;
							fmt.value = b->value[i];
							fwrite(&fmt,sizeof(fmt),1,file);
						}
					}
					fclose(file);
				}
			}
			MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
			if( ret )
			{
				return LIS_ERR_FILE_IO;
			}
		}
	}
	if( !lis_vector_is_null(x) )
	{
		err = 0;
		for(pe=0;pe<nprocs;pe++)
		{
			if( my_rank==pe )
			{
				if( format==LIS_FMT_MM )
				{
					file = fopen(path, "a");
				}
				else
				{
					file = fopen(path, "ab");
				}
				if( file==NULL )
				{
					LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
					err = 1;
				}
				else
				{
					if( format==LIS_FMT_MM )
					{
						for(i=0;i<n;i++)
						{
							fprintf(file, "%d %28.20e\n", is+i+1,x->value[i]);
						}
					}
					else
					{
						for(i=0;i<n;i++)
						{
							fmt.i     = is+i+1;
							fmt.value = x->value[i];
							fwrite(&fmt,sizeof(fmt),1,file);
						}
					}
					fclose(file);
				}
			}
			MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
			if( ret )
			{
				return LIS_ERR_FILE_IO;
			}
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_output_mm_vec(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, FILE *file)
{
	int				n;
	int				i;
	LIS_MM_VECFMT	fmt;

	LIS_DEBUG_FUNC_IN;

	n   = A->n;
	
	if( !lis_vector_is_null(b) )
	{
		if( format==LIS_FMT_MM )
		{
			for(i=0;i<n;i++)
			{
				fprintf(file, "%d %28.20e\n", i+1,b->value[i]);
			}
		}
		else
		{
			for(i=0;i<n;i++)
			{
				fmt.i     = i+1;
				fmt.value = b->value[i];
				fwrite(&fmt,sizeof(fmt),1,file);
			}
		}
	}
	if( !lis_vector_is_null(x) )
	{
		if( format==LIS_FMT_MM )
		{
			for(i=0;i<n;i++)
			{
				fprintf(file, "%d %28.20e\n", i+1,x->value[i]);
			}
		}
		else
		{
			for(i=0;i<n;i++)
			{
				fmt.i     = i+1;
				fmt.value = x->value[i];
				fwrite(&fmt,sizeof(fmt),1,file);
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_output_mm_header"
#ifdef USE_MPI
int lis_output_mm_header(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path, FILE **file)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				err,ret;
	int				isb,isx,endian;

	LIS_DEBUG_FUNC_IN;

	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	isb = 0;
	isx = 0;
	if( !lis_vector_is_null(b) ) isb = 1;
	if( !lis_vector_is_null(x) ) isx = 1;
	endian = 1;
	endian = *(char *)&endian;

	err       = 0;
	MPI_Allreduce(&local_nnz,&nnz,1,MPI_INT,MPI_SUM,A->comm);
	if( my_rank==0 )
	{
		if( format==LIS_FMT_MM )
		{
			*file = fopen(path, "w");
		}
		else
		{
			*file = fopen(path, "wb");
		}
		if( file==NULL )
		{
			LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n",path);
			err = 1;
		}
		else
		{
			fprintf(*file, "%%%%MatrixMarket matrix coordinate real general\n");

			if( format==LIS_FMT_MM )
			{
				if( isb==0 && isx==0 )
				{
					fprintf(*file, "%d %d %d\n", A->gn, A->gn, nnz);
				}
				else
				{
					fprintf(*file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
				}
			}
			else
			{
				fprintf(*file, "%d %d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx, endian+1);
			}
			fclose(*file);
		}
	}
	MPI_Allreduce(&err,&ret,1,MPI_INT,MPI_SUM,A->comm);
	if( ret )
	{
		return LIS_ERR_FILE_IO;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_output_mm_header(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path, FILE **file)
{
	int				nnz;
	int				err;
	int				isb,isx,endian;

	LIS_DEBUG_FUNC_IN;

	nnz = A->nnz;
	isb = 0;
	isx = 0;
	if( !lis_vector_is_null(b) ) isb = 1;
	if( !lis_vector_is_null(x) ) isx = 1;
	endian = 1;
	endian = *(char *)&endian;

	if( format==LIS_FMT_MM )
	{
		*file = fopen(path, "w");
	}
	else
	{
		*file = fopen(path, "wb");
	}
	if( file==NULL )
	{
		LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n",path);
		return LIS_ERR_FILE_IO;
	}
	fprintf(*file, "%%%%MatrixMarket matrix coordinate real general\n");

	if( format==LIS_FMT_MM )
	{
		if( isb==0 && isx==0 )
		{
			fprintf(*file, "%d %d %d\n", A->gn, A->gn, nnz);
		}
		else
		{
			fprintf(*file, "%d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx);
		}
	}
	else
	{
		fprintf(*file, "%d %d %d %d %d %d\n", A->gn, A->gn, nnz, isb, isx, endian+1);
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_output_mm_crs"
#ifdef USE_MPI
int lis_output_mm_crs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,local_nnz,nnz,nprocs,my_rank;
	int				i,j,jj,is;
	int				pe;
	int				err,ret;
	int				isb,isx;
	FILE			*file;
	LIS_MM_MATFMT	fmt;

	LIS_DEBUG_FUNC_IN;

	nprocs    = A->nprocs;
	my_rank   = A->my_rank;
	local_nnz = A->nnz;
	n         = A->n;
	is        = A->is;

	err = lis_output_mm_header(A,b,x,format,path,&file);
	if( err ) return err;


	for(pe=0;pe<nprocs;pe++)
	{
		if( my_rank==pe )
		{
			file = fopen(path, "a");
			if( file==NULL )
			{
				LIS_SETERR1(LIS_ERR_FILE_IO,"cannot open file %s\n", path);
				return LIS_FAILS;
			}
			if( A->matrix_type==LIS_MATRIX_CRS )
			{
				if( format==LIS_FMT_MM )
				{
					for(i=0;i<n;i++)
					{
						for(j=A->ptr[i];j<A->ptr[i+1];j++)
						{
							if( A->index[j]>=n )
							{
								jj = A->l2g_map[A->index[j]-n]+1;
							}
							else
							{
								jj = is + A->index[j]+1;
							}
							fprintf(file, "%d %d %28.20e\n", is+i+1,jj,A->value[j]);
						}
					}
				}
				else
				{
					for(i=0;i<n;i++)
					{
						for(j=A->ptr[i];j<A->ptr[i+1];j++)
						{
							if( A->index[j]>=n )
							{
								jj = A->l2g_map[A->index[j]-n]+1;
							}
							else
							{
								jj = is + A->index[j]+1;
							}
							fmt.i     = is+i+1;
							fmt.j     = jj;
							fmt.value = A->value[j];
							fwrite(&fmt,sizeof(fmt),1,file);
						}
					}
				}
			}
			else
			{
				if( format==LIS_FMT_MM )
				{
					for(i=0;i<n;i++)
					{
						for(j=A->ptr[i];j<A->ptr[i+1];j++)
						{
							if( A->index[j]>=n )
							{
								jj = A->l2g_map[A->index[j]-n]+1;
							}
							else
							{
								jj = is + A->index[j]+1;
							}
							fprintf(file, "%d %d %28.20e\n", jj,is+i+1,A->value[j]);
						}
					}
				}
				else
				{
					for(i=0;i<n;i++)
					{
						for(j=A->ptr[i];j<A->ptr[i+1];j++)
						{
							if( A->index[j]>=n )
							{
								jj = A->l2g_map[A->index[j]-n]+1;
							}
							else
							{
								jj = is + A->index[j]+1;
							}
							fmt.j     = is+i+1;
							fmt.i     = jj;
							fmt.value = A->value[j];
							fwrite(&fmt,sizeof(fmt),1,file);
						}
					}
				}
			}
			fclose(file);
		}
		MPI_Barrier(A->comm);
	}
	err = lis_output_mm_vec(A,b,x,format,path);
	if( err ) return err;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
int lis_output_mm_crs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				n,nnz;
	int				i,j,jj;
	int				err;
	int				isb,isx,endian;
	FILE			*file;
	LIS_MM_MATFMT	fmt;

	LIS_DEBUG_FUNC_IN;

	n   = A->n;
	nnz = A->nnz;

	err = lis_output_mm_header(A,b,x,format,path,&file);
	if( err ) return err;

	if( A->matrix_type==LIS_MATRIX_CRS )
	{
		if( format==LIS_FMT_MM )
		{
			for(i=0;i<n;i++)
			{
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->index[j]+1;
					fprintf(file, "%d %d %28.20e\n", i+1,jj,A->value[j]);
				}
			}
		}
		else
		{
			for(i=0;i<n;i++)
			{
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->index[j]+1;
					fmt.i     = i+1;
					fmt.j     = jj;
					fmt.value = A->value[j];
					fwrite(&fmt,sizeof(fmt),1,file);
				}
			}
		}
	}
	else
	{
		if( format==LIS_FMT_MM )
		{
			for(i=0;i<n;i++)
			{
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->index[j]+1;
					fprintf(file, "%d %d %28.20e\n", jj,i+1,A->value[j]);
				}
			}
		}
		else
		{
			for(i=0;i<n;i++)
			{
				for(j=A->ptr[i];j<A->ptr[i+1];j++)
				{
					jj = A->index[j]+1;
					fmt.j     = i+1;
					fmt.i     = jj;
					fmt.value = A->value[j];
					fwrite(&fmt,sizeof(fmt),1,file);
				}
			}
		}
	}
	lis_output_mm_vec(A,b,x,format,file);
	fclose(file);
	
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif

#undef __FUNC__
#define __FUNC__ "lis_output_mm_ccs"
int lis_output_mm_ccs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int format, char *path)
{
	int				err;

	LIS_DEBUG_FUNC_IN;

	err = lis_output_mm_crs(A,b,x,format,path);

	LIS_DEBUG_FUNC_OUT;
	return err;

}

