/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lis.h"


#undef __FUNC__
#define __FUNC__ "main"
int main(int argc, char* argv[])
{
	LIS_MATRIX		A;
	LIS_VECTOR		x,b,u;
	LIS_SOLVER		solver;
	int				nprocs,my_rank;
	int				np,nsol;
	int				err,iter;
	double			times,itimes,ptimes;
	LIS_REAL		resid,nrm2;
	char			solvername[128];

	LIS_DEBUG_FUNC_IN;


	lis_initialize(argc, argv);

	#ifdef USE_MPI
		MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
		MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
	#else
		nprocs  = 1;
		my_rank = 0;
	#endif

	if( argc < 4 )
	{
		if( my_rank==0 ) printf("test1 matrix_filename result_filename residual_filename [options]\n");
		lis_finalize();
		exit(0);
	}


	#ifdef _OPENMP
		if( my_rank==0 )
		{
			printf("Max Procs   = %d\n",omp_get_num_procs());
			printf("Max Threads = %d\n",omp_get_max_threads());
		}
	#endif

	/* read matrix and vectors from file */
	err = lis_matrix_create(LIS_COMM_WORLD,&A); CHKERR(err);
	err = lis_vector_create(LIS_COMM_WORLD,&b); CHKERR(err);
	err = lis_vector_create(LIS_COMM_WORLD,&x); CHKERR(err);
	err = lis_input(A,b,x,argv[1]);
	CHKERR(err);
	np = A->np;


	err = lis_vector_duplicate(A,&u);
	CHKERR(err);
	if( lis_vector_is_null(b) )
	{
		lis_vector_destroy(b);
		lis_vector_duplicate(u,&b);
		CHKERR(err);
/*		err = lis_vector_set_all(1.0,b);*/
		err = lis_vector_set_all(1.0,u);
		lis_matvec(A,u,b);
	}
	if( lis_vector_is_null(x) )
	{
		lis_vector_destroy(x);
		err = lis_vector_duplicate(u,&x);
		CHKERR(err);
	}

	err = lis_solver_create(&solver); CHKERR(err);
	lis_solver_set_option("-print mem",solver);
	lis_solver_set_optionC(solver);

	err = lis_solve(A,b,x,solver);
	CHKERR(err);
	lis_solver_get_iters(solver,&iter);
	lis_solver_get_times(solver,&times,&itimes,&ptimes);
	lis_solver_get_residualnorm(solver,&resid);
	lis_solver_get_solver(solver,&nsol);
	lis_get_solvername(nsol,solvername);
	
	lis_vector_axpy(-1.0,x,u);
	lis_vector_nrm2(u,&nrm2);
	if( my_rank==0 )
	{
		printf("%s: iter = %d iter_double = %d time = %e (p=%e i=%e)\n",solvername,iter, solver->iter2, times, ptimes, itimes);
		printf("%s: Last Residual = %e \n",solvername,resid);
	}

	/* write solution */
	lis_output_vector(x,LIS_FMT_MM,argv[2]);

	/* write residual */
	lis_output_residual_history(solver, argv[3]);

	lis_solver_destroy(solver);
	lis_matrix_destroy(A);
	lis_vector_destroy(b);
	lis_vector_destroy(x);
	lis_vector_destroy(u);

	lis_finalize();

	LIS_DEBUG_FUNC_OUT;
	return 0;
}

