      implicit none
      
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif
#include "lisf.h"

#ifdef USE_MPI
      include 'mpif.h'
#endif

      LIS_MATRIX    A
      LIS_VECTOR    b,x,u
      LIS_SOLVER    solver
      integer      ierr
      integer      my_rank,nprocs
      integer      n,gn,i
      integer      nsol,iter
      real*8       times,itimes,ptimes
      LIS_REAL      resid
      character*256 fname,solname,resname
      character*10 solvername
      integer*4 iargc
      
      call lis_initialize(ierr) 

#ifdef USE_MPI
      call MPI_Comm_size(MPI_COMM_WORLD,nprocs,ierr)
      call MPI_Comm_rank(MPI_COMM_WORLD,my_rank,ierr)
#else
      nprocs  = 1
      my_rank = 0
#endif

      i = iargc()
      if( i.lt.3 ) then
        if( my_rank.eq.0 ) then
          write(*,*) 'test1f matrix_filename result_filename ',
     .              'residual_filename [options]'
          call lis_finalize(ierr)
        endif
        stop
      endif

! read matrix and vectors from file 
      call getarg(1,fname)
      call lis_matrix_create(LIS_COMM_WORLD,A,ierr)
      call CHKERR(ierr)
      call lis_vector_create(LIS_COMM_WORLD,b,ierr)
      call CHKERR(ierr)
      call lis_vector_create(LIS_COMM_WORLD,x,ierr)
      call CHKERR(ierr)
      call lis_matrix_set_type(A,LIS_MATRIX_CRS,ierr)
      call lis_input(A,b,x,fname,ierr)
	call CHKERR(ierr);
	call lis_matrix_get_size(A,n,gn,ierr)

      call lis_vector_duplicate(A,u,ierr)
      call CHKERR(ierr)
      call lis_is_vector_null(b,ierr)
      if( ierr.eq.LIS_TRUE ) then
        call lis_vector_destroy(b,ierr)
        call lis_vector_duplicate(u,b,ierr)
        call CHKERR(ierr)
        call lis_vector_set_all(1.0d0,u,ierr)
        call lis_matvec(A,u,b,ierr)
      endif

      call lis_is_vector_null(x,ierr)
      if( ierr.eq.LIS_TRUE ) then
        call lis_vector_destroy(x,ierr)
        call lis_vector_duplicate(u,x,ierr)
        call CHKERR(ierr)
      endif

      call lis_solver_create(solver,ierr)
      call CHKERR(ierr)
      call lis_solver_set_option('-print mem',solver,ierr)
      call lis_solver_set_optionC(solver,ierr)
      
      call lis_solve(A,b,x,solver,ierr)
      call CHKERR(ierr)

	call lis_solver_get_iters(solver,iter,ierr)
	call lis_solver_get_times(solver,times,itimes,ptimes,ierr)
	call lis_solver_get_residualnorm(solver,resid,ierr)
      call lis_solver_get_solver(solver,nsol,ierr)
      call lis_get_solvername(nsol,solvername,ierr)
	
      if( my_rank.eq.0 ) then
        write(*,*) solvername,': iter = ',iter,' time = ',times,
     .              ' (p=',ptimes,' i=',itimes,')'
        write(*,*) solvername,': Last Residual = ',resid
      endif

! write solution 
      call getarg(2,solname)
	call lis_output_vector(x,LIS_FMT_MM,solname,ierr);

! write residual 
      call getarg(3,resname)
	call lis_output_residual_history(solver, resname,ierr);

	call lis_solver_destroy(solver,ierr)
      call lis_vector_destroy(u,ierr)
      call lis_vector_destroy(x,ierr)
      call lis_vector_destroy(b,ierr)
      call lis_matrix_destroy(A,ierr)

      call lis_finalize(ierr)

      stop
      end
      
