/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lis.h"


#undef __FUNC__
#define __FUNC__ "main"
int main(int argc, char* argv[])
{
	LIS_MATRIX		A;
	LIS_VECTOR		x,b,u;
	LIS_SOLVER		solver;
	int				i,k,n,nnz,m,nn,gn,ii,j,jj;
	int				is,ie;
	int				nprocs,my_rank;
	int				np,nsol;
	int				err,iter,mtype;
	double			times,itimes,ptimes,flops;
	LIS_REAL		resid;
	char			solvername[128];
	int				*ptr,*index;
	LIS_SCALAR		*value,gamma;
	FILE			*file;

	LIS_DEBUG_FUNC_IN;


	lis_initialize(argc, argv);

	#ifdef USE_MPI
		MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
		MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
	#else
		nprocs  = 1;
		my_rank = 0;
	#endif

	if( argc < 3 )
	{
		if( my_rank==0 ) printf("%s n gamma [options]\n",argv[0]);
		lis_finalize();
		exit(0);
	}

	gn  = atoi(argv[1]);
	gamma  = atof(argv[2]);
	if( gn<=0 )
	{
		if( my_rank==0 ) printf("n=%d <=0 \n",gn);
		lis_finalize();
		exit(0);
	}

		
	if(my_rank==0) printf("n=%d gamma=%f\n",gn,gamma);

	#ifdef _OPENMP
		printf("Max Procs   = %d\n",omp_get_num_procs());
		printf("Max Threads = %d\n",omp_get_max_threads());
	#endif

	/* create matrix and vectors */
    err = lis_matrix_create(LIS_COMM_WORLD,&A); CHKERR(err);
    err = lis_matrix_set_size(A,0,gn); CHKERR(err);
	err = lis_matrix_get_size(A,&n,&gn); CHKERR(err);
	err = lis_matrix_malloc_crs(n,3*n,&ptr,&index,&value); CHKERR(err);
	err = lis_matrix_get_range(A,&is,&ie); CHKERR(err);

	k = 0;
	ptr[0] = 0;
	for(ii=is;ii<ie;ii++)
	{
		if( ii>1 )    { jj = ii - 2; index[k] = jj; value[k++] = gamma;}
		if( ii<gn-1 ) { jj = ii + 1; index[k] = jj; value[k++] = 1.0;}
		index[k] = ii; value[k++] = 2.0;
		ptr[ii-is+1] = k;
	}
	err = lis_matrix_set_crs(ptr[ie-is],ptr,index,value,A); CHKERR(err);
	err = lis_matrix_assemble(A); CHKERR(err);

	err = lis_vector_duplicate(A,&u); CHKERR(err);
	err = lis_vector_duplicate(u,&b); CHKERR(err);
	err = lis_vector_duplicate(u,&x); CHKERR(err);

	err = lis_vector_set_all(1.0,b);
	err = lis_vector_set_all(1.0,x);

	err = lis_solver_create(&solver); CHKERR(err);
	lis_solver_set_optionC(solver);

	err = lis_solve(A,b,x,solver); CHKERR(err);
	lis_solver_get_iters(solver,&iter);
	lis_solver_get_times(solver,&times,&itimes,&ptimes);
	lis_solver_get_residualnorm(solver,&resid);
	lis_solver_get_solver(solver,&nsol);
	lis_get_solvername(nsol,solvername);
	if( my_rank==0 )
	{
		printf("%s: iter = %d iter_double = %d time = %e (p=%e i=%e)\n",solvername,iter, solver->iter2, times, ptimes, itimes);
		printf("%s: Last Residual = %e \n",solvername,resid);
	}
	
	lis_solver_destroy(solver);
	lis_matrix_destroy(A);
	lis_vector_destroy(b);
	lis_vector_destroy(x);
	lis_vector_destroy(u);

	lis_finalize();

	LIS_DEBUG_FUNC_OUT;
	return 0;
}


