/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lis.h"


#undef __FUNC__
#define __FUNC__ "main"
int main(int argc, char* argv[])
{
	LIS_MATRIX		A;
	LIS_VECTOR		b,x;
	int				nprocs,my_rank;
	int				format;
	int				err;

	LIS_DEBUG_FUNC_IN;


	lis_initialize(argc, argv);

	#ifdef USE_MPI
		MPI_Comm_size(MPI_COMM_WORLD,&nprocs);
		MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
	#else
		nprocs  = 1;
		my_rank = 0;
	#endif

	if( argc < 4 )
	{
err_print:
		if( my_rank==0 ) printf("test5 out_format out_matrix_filename in_matrix_filename [in_rhs_file in_init_file]\n");
		lis_finalize();
		exit(0);
	}


	if( strncmp(argv[1], "mmb", 3)==0 )
	{
		format = LIS_FMT_MMB;
	}
	else if( strncmp(argv[1], "mm", 2)==0 )
	{
		format = LIS_FMT_MM;
	}
	else
	{
		goto err_print;
	}

	/* read matrix and vectors from file */
	err = lis_matrix_create(LIS_COMM_WORLD,&A); CHKERR(err);
	err = lis_vector_create(LIS_COMM_WORLD,&b); CHKERR(err);
	err = lis_vector_create(LIS_COMM_WORLD,&x); CHKERR(err);
	err = lis_input(A,b,x,argv[3]);
	CHKERR(err);
	if( argc>4 )
	{
		lis_vector_destroy(b);
		err = lis_input_vector(b,argv[4],LIS_COMM_WORLD);
		CHKERR(err);
		if( A->n!=b->n )
		{
			printf("different dimension: matrix A=%d, vector b=%d\n",A->n,b->n);
			goto err_to;
		}
		if( argc==6 )
		{
			lis_vector_destroy(x);
			err = lis_input_vector(x,argv[5],LIS_COMM_WORLD);
			CHKERR(err);
			if( A->n!=x->n )
			{
				printf("different dimension: matrix A=%d, vector x=%d\n",A->n,x->n);
				goto err_to;
			}
		}
	}

	err = lis_output(A,b,x,format,argv[2]);
	CHKERR(err);

err_to:
	lis_matrix_destroy(A);
	lis_vector_destroy(b);
	lis_vector_destroy(x);

	lis_finalize();

	LIS_DEBUG_FUNC_OUT;
	return 0;
}

